# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `rational_quadratic_spline.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import rational_quadratic_spline
import jax
import jax.numpy as jnp
import numpy as np


def _make_bijector(params_shape,
                   zero_params=False,
                   num_bins=8,
                   range_min=0.,
                   range_max=1.,
                   boundary_slopes='unconstrained'):
  """Returns RationalQuadraticSpline bijector."""
  params_shape += (3 * num_bins + 1,)
  if zero_params:
    params = jnp.zeros(params_shape)
  else:
    key = jax.random.PRNGKey(101)
    params = jax.random.normal(key, params_shape)
  return rational_quadratic_spline.RationalQuadraticSpline(
      params,
      range_min=range_min,
      range_max=range_max,
      boundary_slopes=boundary_slopes)


class RationalQuadraticSplineTest(parameterized.TestCase):
  """Tests for rational quadratic spline."""

  def test_properties(self):
    bijector = _make_bijector(params_shape=(4, 5), num_bins=8)
    self.assertEqual(bijector.is_constant_jacobian, False)
    self.assertEqual(bijector.is_constant_log_det, False)
    assert bijector.num_bins == 8
    self.assertEqual(bijector.knot_slopes.shape, (4, 5, 9))
    self.assertEqual(bijector.x_pos.shape, (4, 5, 9))
    self.assertEqual(bijector.y_pos.shape, (4, 5, 9))

  @parameterized.named_parameters(
      ('params.shape[-1] < 4',
       {'params': np.zeros((3,)), 'range_min': 0., 'range_max': 1.}),
      ('params.shape[-1] % 3 is not 1',
       {'params': np.zeros((8,)), 'range_min': 0., 'range_max': 1.}),
      ('inconsistent range min and max',
       {'params': np.zeros((10,)), 'range_min': 1., 'range_max': 0.9}),
      ('negative min_bin_size',
       {'params': np.zeros((10,)), 'range_min': 0., 'range_max': 1.,
        'min_bin_size': -0.1}),
      ('negative min_knot_slope',
       {'params': np.zeros((10,)), 'range_min': 0., 'range_max': 1.,
        'min_knot_slope': -0.1}),
      ('min_knot_slope above 1',
       {'params': np.zeros((10,)), 'range_min': 0., 'range_max': 1.,
        'min_knot_slope': 1.3}),
      ('invalid boundary_slopes',
       {'params': np.zeros((10,)), 'range_min': 0., 'range_max': 1.,
        'boundary_slopes': 'invalid_value'}),
      ('num_bins * min_bin_size greater than total_size',
       {'params': np.zeros((10,)), 'range_min': 0., 'range_max': 1.,
        'min_bin_size': 0.9}),
  )
  def test_invalid_properties(self, bij_params):
    with self.assertRaises(ValueError):
      rational_quadratic_spline.RationalQuadraticSpline(**bij_params)

  @chex.all_variants
  def test_shapes_are_correct(self):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bijector = _make_bijector(params_shape=(4, 5))
    # Forward methods.
    y, logdet = self.variant(bijector.forward_and_log_det)(x)
    self.assertEqual(y.shape, (2, 3, 4, 5))
    self.assertEqual(logdet.shape, (2, 3, 4, 5))
    # Inverse methods.
    x, logdet = self.variant(bijector.inverse_and_log_det)(y)
    self.assertEqual(x.shape, (2, 3, 4, 5))
    self.assertEqual(logdet.shape, (2, 3, 4, 5))

  @chex.all_variants
  def test_broadcasting_is_correct(self):
    z = 0.5 * jnp.ones((2, 2))
    # Broadcast along first axis.
    bijector = _make_bijector(params_shape=(2,))
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(z)
    x, logdet_inv = self.variant(bijector.inverse_and_log_det)(z)
    np.testing.assert_array_equal(y[0], y[1])
    np.testing.assert_array_equal(x[0], x[1])
    np.testing.assert_array_equal(logdet_fwd[0], logdet_fwd[1])
    np.testing.assert_array_equal(logdet_inv[0], logdet_inv[1])
    self.assertFalse(jnp.allclose(y[:, 0], y[:, 1]))
    self.assertFalse(jnp.allclose(x[:, 0], x[:, 1]))
    self.assertFalse(jnp.allclose(logdet_fwd[:, 0], logdet_fwd[:, 1]))
    self.assertFalse(jnp.allclose(logdet_inv[:, 0], logdet_inv[:, 1]))
    # Broadcast along second axis.
    bijector = _make_bijector(params_shape=(2, 1))
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(z)
    x, logdet_inv = self.variant(bijector.inverse_and_log_det)(z)
    np.testing.assert_array_equal(y[:, 0], y[:, 1])
    np.testing.assert_array_equal(x[:, 0], x[:, 1])
    np.testing.assert_array_equal(logdet_fwd[:, 0], logdet_fwd[:, 1])
    np.testing.assert_array_equal(logdet_inv[:, 0], logdet_inv[:, 1])
    self.assertFalse(jnp.allclose(y[0], y[1]))
    self.assertFalse(jnp.allclose(x[0], x[1]))
    self.assertFalse(jnp.allclose(logdet_fwd[0], logdet_fwd[1]))
    self.assertFalse(jnp.allclose(logdet_inv[0], logdet_inv[1]))

  @chex.all_variants
  def test_is_identity_for_zero_params(self):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bijector = _make_bijector(params_shape=(4, 5), zero_params=True)
    # Forward methods.
    y, logdet = self.variant(bijector.forward_and_log_det)(x)
    np.testing.assert_allclose(y, x, atol=5e-5)
    np.testing.assert_allclose(logdet, jnp.zeros((2, 3, 4, 5)), atol=5e-5)
    # Inverse methods.
    x, logdet = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(y, x, atol=5e-5)
    np.testing.assert_allclose(logdet, jnp.zeros((2, 3, 4, 5)), atol=5e-5)

  @chex.all_variants
  def test_inverse_methods_are_correct(self):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bijector = _make_bijector(params_shape=(4, 5))
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(x)
    x_rec, logdet_inv = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(x_rec, x, atol=7e-4)
    np.testing.assert_allclose(logdet_fwd, -logdet_inv, atol=7e-4)

  @chex.all_variants
  def test_is_monotonically_increasing(self):
    z = jnp.linspace(start=-2, stop=2, num=100)
    bijector = _make_bijector(params_shape=())
    y = self.variant(bijector.forward)(z)
    x = self.variant(bijector.inverse)(z)
    np.testing.assert_array_less(y[:-1], y[1:])
    np.testing.assert_array_less(x[:-1], x[1:])

  @chex.all_variants
  def test_composite_methods_are_consistent(self):
    key = jax.random.PRNGKey(42)
    bijector = _make_bijector(params_shape=(4, 5))
    # Forward methods.
    x = jax.random.normal(key, (2, 3, 4, 5))
    y1 = self.variant(bijector.forward)(x)
    logdet1 = self.variant(bijector.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bijector.forward_and_log_det)(x)
    np.testing.assert_allclose(y1, y2, atol=1e-6)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-6)
    # Inverse methods.
    y = jax.random.normal(key, (2, 3, 4, 5))
    x1 = self.variant(bijector.inverse)(y)
    logdet1 = self.variant(bijector.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(x1, x2, atol=1e-6)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-6)

  @chex.all_variants
  def test_boundary_conditions(self):
    a = jnp.array(0.)
    b = jnp.array(1.)
    # Unconstrained boundary slopes.
    bijector = _make_bijector(
        params_shape=(),
        range_min=float(a),
        range_max=float(b),
        boundary_slopes='unconstrained')
    log_slope_a = self.variant(bijector.forward_log_det_jacobian)(a)
    log_slope_b = self.variant(bijector.forward_log_det_jacobian)(b)
    self.assertEqual(self.variant(bijector.forward)(a), a)
    self.assertEqual(self.variant(bijector.forward)(b), b)
    self.assertFalse(jnp.allclose(log_slope_a, 0.))
    self.assertFalse(jnp.allclose(log_slope_b, 0.))
    # Lower boundary slope equal to 1.
    bijector = _make_bijector(
        params_shape=(),
        range_min=float(a),
        range_max=float(b),
        boundary_slopes='lower_identity')
    log_slope_a = self.variant(bijector.forward_log_det_jacobian)(a)
    log_slope_b = self.variant(bijector.forward_log_det_jacobian)(b)
    self.assertEqual(self.variant(bijector.forward)(a), a)
    self.assertEqual(self.variant(bijector.forward)(b), b)
    self.assertEqual(log_slope_a, 0.)
    self.assertFalse(jnp.allclose(log_slope_b, 0.))
    # Upper boundary slope equal to 1.
    bijector = _make_bijector(
        params_shape=(),
        range_min=float(a),
        range_max=float(b),
        boundary_slopes='upper_identity')
    log_slope_a = self.variant(bijector.forward_log_det_jacobian)(a)
    log_slope_b = self.variant(bijector.forward_log_det_jacobian)(b)
    self.assertEqual(self.variant(bijector.forward)(a), a)
    self.assertEqual(self.variant(bijector.forward)(b), b)
    self.assertFalse(jnp.allclose(log_slope_a, 0.))
    self.assertEqual(log_slope_b, 0.)
    # Both boundary slopes equal to 1.
    bijector = _make_bijector(
        params_shape=(),
        range_min=float(a),
        range_max=float(b),
        boundary_slopes='identity')
    log_slope_a = self.variant(bijector.forward_log_det_jacobian)(a)
    log_slope_b = self.variant(bijector.forward_log_det_jacobian)(b)
    self.assertEqual(self.variant(bijector.forward)(a), a)
    self.assertEqual(self.variant(bijector.forward)(b), b)
    self.assertEqual(log_slope_a, 0.)
    self.assertEqual(log_slope_b, 0.)
    # Circular spline (periodic slope).
    bijector = _make_bijector(
        params_shape=(),
        range_min=float(a),
        range_max=float(b),
        boundary_slopes='circular')
    log_slope_a = self.variant(bijector.forward_log_det_jacobian)(a)
    log_slope_b = self.variant(bijector.forward_log_det_jacobian)(b)
    self.assertEqual(self.variant(bijector.forward)(a), a)
    self.assertEqual(self.variant(bijector.forward)(b), b)
    self.assertEqual(log_slope_a, log_slope_b)
    self.assertFalse(jnp.allclose(log_slope_b, 0.))

  @chex.all_variants
  @parameterized.parameters(
      ((3, 4), (3, 4)),
      ((3, 4), (3, 1)),
      ((3, 4), (4,)),
      ((3, 4), ()),
      ((3, 1), (3, 4)),
      ((4,), (3, 4)),
      ((), (3, 4)),
  )
  def test_batched_parameters(self, params_batch_shape, input_batch_shape):
    k1, k2 = jax.random.split(jax.random.PRNGKey(42), 2)
    num_bins = 4
    param_dim = 3 * num_bins + 1
    params = jax.random.normal(k1, params_batch_shape + (param_dim,))
    bijector = rational_quadratic_spline.RationalQuadraticSpline(
        params, range_min=0., range_max=1.)

    x = jax.random.uniform(k2, input_batch_shape)
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(x)
    z, logdet_inv = self.variant(bijector.inverse_and_log_det)(x)

    output_batch_shape = jnp.broadcast_arrays(params[..., 0], x)[0].shape

    self.assertEqual(y.shape, output_batch_shape)
    self.assertEqual(z.shape, output_batch_shape)
    self.assertEqual(logdet_fwd.shape, output_batch_shape)
    self.assertEqual(logdet_inv.shape, output_batch_shape)

    params = jnp.broadcast_to(
        params, output_batch_shape + (param_dim,)).reshape((-1, param_dim))
    x = jnp.broadcast_to(x, output_batch_shape).flatten()
    y = y.flatten()
    z = z.flatten()
    logdet_fwd = logdet_fwd.flatten()
    logdet_inv = logdet_inv.flatten()

    for i in range(np.prod(output_batch_shape)):
      bijector = rational_quadratic_spline.RationalQuadraticSpline(
          params[i], range_min=0., range_max=1.)
      this_y, this_logdet_fwd = self.variant(bijector.forward_and_log_det)(x[i])
      this_z, this_logdet_inv = self.variant(bijector.inverse_and_log_det)(x[i])
      np.testing.assert_allclose(this_y, y[i], atol=1e-7)
      np.testing.assert_allclose(this_z, z[i], atol=1e-6)
      np.testing.assert_allclose(this_logdet_fwd, logdet_fwd[i], atol=1e-5)
      np.testing.assert_allclose(this_logdet_inv, logdet_inv[i], atol=1e-5)

  @chex.all_variants
  @parameterized.parameters(
      (-1., 4., -3., 1.,),  # when b >= 0
      (1., -4., 3., 3.),  # when b < 0
      (-1., 2., -1., 1.),  # when b**2 - 4*a*c = 0, and b >= 0
      (1., -2., 1., 1.),  # when b**2 - 4*a*c = 0, and b < 0
  )
  def test_safe_quadratic_root(self, a, b, c, x):
    a = jnp.array(a)
    b = jnp.array(b)
    c = jnp.array(c)
    x = jnp.array(x)
    sol_x, grad = self.variant(jax.value_and_grad(
        rational_quadratic_spline._safe_quadratic_root))(a, b, c)
    np.testing.assert_allclose(sol_x, x, atol=1e-5)
    self.assertFalse(np.any(np.isnan(grad)))

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = _make_bijector(params_shape=(4, 5))
    x = np.zeros((2, 3, 4, 5))
    f(x, bijector)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Quantized distribution."""

from typing import cast, Optional, Tuple

import chex
from distrax._src.distributions import distribution as base_distribution
from distrax._src.utils import conversion
from distrax._src.utils import math
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
DistributionLike = base_distribution.DistributionLike
DistributionT = base_distribution.DistributionT
EventT = base_distribution.EventT


class Quantized(
    base_distribution.Distribution[Array, Tuple[int, ...], jnp.dtype],):
  """Distribution representing the quantization `Y = ceil(X)`.

  Given an input distribution `p(x)` over a univariate random variable `X`,
  sampling from a `Quantized` distribution amounts to sampling `x ~ p(x)` and
  then setting `y = ceil(x)`. The returned samples are integer-valued and of the
  same `dtype` as the base distribution.
  """

  equiv_tfp_cls = tfd.QuantizedDistribution

  def __init__(self,
               distribution: DistributionLike,
               low: Optional[Numeric] = None,
               high: Optional[Numeric] = None,
               eps: Optional[Numeric] = None):
    """Initializes a Quantized distribution.

    Args:
      distribution: The base distribution to be quantized.
      low: Lowest possible quantized value, such that samples are `y >=
        ceil(low)`. Its shape must broadcast with the shape of samples from
        `distribution` and must not result in additional batch dimensions after
        broadcasting.
      high: Highest possible quantized value, such that samples are `y <=
        floor(high)`. Its shape must broadcast with the shape of samples from
        `distribution` and must not result in additional batch dimensions after
        broadcasting.
      eps: An optional gap to enforce between "big" and "small". Useful for
        avoiding NANs in computing log_probs, when "big" and "small"
        are too close.
    """
    self._dist: base_distribution.Distribution[Array, Tuple[
        int, ...], jnp.dtype] = conversion.as_distribution(distribution)
    self._eps = eps
    if self._dist.event_shape:
      raise ValueError(f'The base distribution must be univariate, but its '
                       f'`event_shape` is {self._dist.event_shape}.')
    dtype = self._dist.dtype
    if low is None:
      self._low = None
    else:
      self._low = jnp.asarray(jnp.ceil(low), dtype=dtype)
      if len(self._low.shape) > len(self._dist.batch_shape):
        raise ValueError('The parameter `low` must not result in additional '
                         'batch dimensions.')
    if high is None:
      self._high = None
    else:
      self._high = jnp.asarray(jnp.floor(high), dtype=dtype)
      if len(self._high.shape) > len(self._dist.batch_shape):
        raise ValueError('The parameter `high` must not result in additional '
                         'batch dimensions.')
    super().__init__()

  @property
  def distribution(
      self
  ) -> base_distribution.Distribution[Array, Tuple[int, ...], jnp.dtype]:
    """Base distribution `p(x)`."""
    return self._dist

  @property
  def low(self) -> Optional[Array]:
    """Lowest value that quantization returns."""
    if self._low is None:
      return None
    return jnp.broadcast_to(self._low, self.batch_shape + self.event_shape)

  @property
  def high(self) -> Optional[Array]:
    """Highest value that quantization returns."""
    if self._high is None:
      return None
    return jnp.broadcast_to(self._high, self.batch_shape + self.event_shape)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    event_shape = self.distribution.event_shape
    # TODO(b/149413467): Remove explicit casting when resolved.
    return cast(Tuple[int, ...], event_shape)

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self.distribution.batch_shape

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    samples = self.distribution.sample(seed=key, sample_shape=n)
    samples = jnp.ceil(samples)

    # Apply overflow and underflow conditions.
    if self.low is not None:
      samples = jnp.where(samples < self.low, self.low, samples)
    if self.high is not None:
      samples = jnp.where(samples > self.high, self.high, samples)

    return samples

  def _sample_n_and_log_prob(self, key: PRNGKey, n: int) -> Tuple[Array, Array]:
    """See `Distribution._sample_n_and_log_prob`."""
    samples = self._sample_n(key, n)
    log_cdf = self.distribution.log_cdf(samples)
    log_cdf_m1 = self.distribution.log_cdf(samples - 1.)
    log_sf = self.distribution.log_survival_function(samples)
    log_sf_m1 = self.distribution.log_survival_function(samples - 1.)
    if self.high is not None:
      # `samples - 1.` is definitely lower than `high`.
      log_cdf = jnp.where(samples < self.high, log_cdf, 0.)
      log_sf = jnp.where(samples < self.high, log_sf, -jnp.inf)
    if self.low is not None:
      # `samples` is definitely greater than or equal to `low`.
      log_cdf_m1 = jnp.where(samples - 1. < self.low, -jnp.inf, log_cdf_m1)
      log_sf_m1 = jnp.where(samples - 1. < self.low, 0., log_sf_m1)
    # Use the survival function instead of the CDF when its value is smaller,
    # which happens to the right of the median of the distribution.
    big = jnp.where(log_sf < log_cdf, log_sf_m1, log_cdf)
    small = jnp.where(log_sf < log_cdf, log_sf, log_cdf_m1)
    log_probs = math.log_expbig_minus_expsmall(big, small)
    return samples, log_probs

  def log_prob(self, value: EventT) -> Array:
    """Calculates the log probability of an event.

    This implementation differs slightly from the one in TFP, as it returns
    `-jnp.inf` on non-integer values instead of returning the log prob of the
    floor of the input. In addition, this implementation also returns `-jnp.inf`
    on inputs that are outside the support of the distribution (as opposed to
    `nan`, like TFP does). On other integer values, both implementations are
    identical.

    Similar to TFP, the log prob is computed using either the CDF or the
    survival function to improve numerical stability. With infinite precision
    the two computations would be equal.

    Args:
      value: An event.

    Returns:
      The log probability log P(value).
    """
    log_cdf = self.log_cdf(value)
    log_cdf_m1 = self.log_cdf(value - 1.)
    log_sf = self.log_survival_function(value)
    log_sf_m1 = self.log_survival_function(value - 1.)
    # Use the survival function instead of the CDF when its value is smaller,
    # which happens to the right of the median of the distribution.
    big = jnp.where(log_sf < log_cdf, log_sf_m1, log_cdf)
    small = jnp.where(log_sf < log_cdf, log_sf, log_cdf_m1)
    if self._eps is not None:
      # use stop_gradient to block updating in this case
      big = jnp.where(big - small > self._eps, big,
                      jax.lax.stop_gradient(small) + self._eps)
    log_probs = math.log_expbig_minus_expsmall(big, small)

    # Return -inf when evaluating on non-integer value.
    is_integer = jnp.where(value > jnp.floor(value), False, True)
    log_probs = jnp.where(is_integer, log_probs, -jnp.inf)

    # Return -inf and not NaN when outside of [low, high].
    # If the CDF is used, `value > high` is already treated correctly;
    # to fix the return value for `value < low` we test whether `log_cdf` is
    # finite; `log_sf_m1` will be `0.` in this regime.
    # If the survival function is used the reverse case applies; to fix the
    # case `value > high` we test whether `log_sf_m1` is finite; `log_cdf` will
    # be `0.` in this regime.
    is_outside = jnp.logical_or(jnp.isinf(log_cdf), jnp.isinf(log_sf_m1))
    log_probs = jnp.where(is_outside, -jnp.inf, log_probs)

    return log_probs

  def prob(self, value: EventT) -> Array:
    """Calculates the probability of an event.

    This implementation differs slightly from the one in TFP, as it returns 0
    on non-integer values instead of returning the prob of the floor of the
    input. It is identical for integer values.

    Similar to TFP, the probability is computed using either the CDF or the
    survival function to improve numerical stability. With infinite precision
    the two computations would be equal.

    Args:
      value: An event.

    Returns:
      The probability P(value).
    """
    cdf = self.cdf(value)
    cdf_m1 = self.cdf(value - 1.)
    sf = self.survival_function(value)
    sf_m1 = self.survival_function(value - 1.)
    # Use the survival function instead of the CDF when its value is smaller,
    # which happens to the right of the median of the distribution.
    probs = jnp.where(sf < cdf, sf_m1 - sf, cdf - cdf_m1)

    # Return 0. when evaluating on non-integer value.
    is_integer = jnp.where(value > jnp.floor(value), False, True)
    probs = jnp.where(is_integer, probs, 0.)
    return probs

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    # The log CDF of a quantized distribution is piecewise constant on half-open
    # intervals:
    #    ... [n-2   n-1) [n-1   n) [n   n+1) [n+1   n+2) ...
    # with log CDF(n) <= log CDF(n+1), because the distribution only has mass on
    # integer values. Therefore: P[Y <= value] = P[Y <= floor(value)].
    y = jnp.floor(value)
    result = self.distribution.log_cdf(y)
    # Update result outside of the interval [low, high].
    if self.low is not None:
      result = jnp.where(y < self.low, -jnp.inf, result)
    if self.high is not None:
      result = jnp.where(y < self.high, result, 0.)
    return result

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    # The CDF of a quantized distribution is piecewise constant on half-open
    # intervals:
    #    ... [n-2   n-1) [n-1   n) [n   n+1) [n+1   n+2) ...
    # with CDF(n) <= CDF(n+1), because the distribution only has mass on integer
    # values. Therefore: P[Y <= value] = P[Y <= floor(value)].
    y = jnp.floor(value)
    result = self.distribution.cdf(y)
    # Update result outside of the interval [low, high].
    if self.low is not None:
      result = jnp.where(y < self.low, 0., result)
    if self.high is not None:
      result = jnp.where(y < self.high, result, 1.)
    return result

  def log_survival_function(self, value: EventT) -> Array:
    """Calculates the log of the survival function of an event.

    This implementation differs slightly from TFP, in that it returns the
    correct log of the survival function for non-integer values, that is, it
    always equates to `log(1 - CDF(value))`. It is identical for integer values.

    Args:
      value: An event.

    Returns:
      The log of the survival function `log P[Y > value]`.
    """
    # The log of the survival function of a quantized distribution is piecewise
    # constant on half-open intervals:
    #    ... [n-2   n-1) [n-1   n) [n   n+1) [n+1   n+2) ...
    # with log sf(n) >= log sf(n+1), because the distribution only has mass on
    # integer values. Therefore: log P[Y > value] = log P[Y > floor(value)].
    y = jnp.floor(value)
    result = self.distribution.log_survival_function(y)
    # Update result outside of the interval [low, high].
    if self._low is not None:
      result = jnp.where(y < self._low, 0., result)
    if self._high is not None:
      result = jnp.where(y < self._high, result, -jnp.inf)
    return result

  def survival_function(self, value: EventT) -> Array:
    """Calculates the survival function of an event.

    This implementation differs slightly from TFP, in that it returns the
    correct survival function for non-integer values, that is, it always
    equates to `1 - CDF(value)`. It is identical for integer values.

    Args:
      value: An event.

    Returns:
      The survival function `P[Y > value]`.
    """
    # The survival function of a quantized distribution is piecewise
    # constant on half-open intervals:
    #    ... [n-2   n-1) [n-1   n) [n   n+1) [n+1   n+2) ...
    # with sf(n) >= sf(n+1), because the distribution only has mass on
    # integer values. Therefore: P[Y > value] = P[Y > floor(value)].
    y = jnp.floor(value)
    result = self.distribution.survival_function(y)
    # Update result outside of the interval [low, high].
    if self._low is not None:
      result = jnp.where(y < self._low, 1., result)
    if self._high is not None:
      result = jnp.where(y < self._high, result, 0.)
    return result

  def __getitem__(self, index) -> 'Quantized':
    """See `Distribution.__getitem__`."""
    index = base_distribution.to_batch_shape_index(self.batch_shape, index)
    low = None if self._low is None else self.low[index]
    high = None if self._high is None else self.high[index]
    return Quantized(distribution=self.distribution[index], low=low, high=high)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `gamma.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import gamma
from distrax._src.utils import equivalence
import jax.experimental
import jax.numpy as jnp
import numpy as np


class GammaTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(gamma.Gamma)

  @parameterized.named_parameters(
      ('0d params', (), (), ()),
      ('1d params', (2,), (2,), (2,)),
      ('2d params, no broadcast', (3, 2), (3, 2), (3, 2)),
      ('2d params, broadcasted concentration', (2,), (3, 2), (3, 2)),
      ('2d params, broadcasted rate', (3, 2), (2,), (3, 2)),
  )
  def test_properties(self, concentration_shape, rate_shape, batch_shape):
    rng = np.random.default_rng(42)
    concentration = 0.1 + rng.uniform(size=concentration_shape)
    rate = 0.1 + rng.uniform(size=rate_shape)
    dist = gamma.Gamma(concentration, rate)
    self.assertEqual(dist.event_shape, ())
    self.assertEqual(dist.batch_shape, batch_shape)
    self.assertion_fn(rtol=2e-2)(
        dist.concentration, np.broadcast_to(concentration, batch_shape))
    self.assertion_fn(rtol=2e-2)(dist.rate, np.broadcast_to(rate, batch_shape))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std gamma, no shape', (1, 1), ()),
      ('1d std gamma, int shape', (1, 1), 1),
      ('1d std gamma, 1-tuple shape', (1, 1), (1,)),
      ('1d std gamma, 2-tuple shape', (1, 1), (2, 2)),
      ('2d std gamma, no shape', (np.ones(2), np.ones(2)), ()),
      ('2d std gamma, int shape', ([1, 1], [1, 1]), 1),
      ('2d std gamma, 1-tuple shape', (np.ones(2), np.ones(2)), (1,)),
      ('2d std gamma, 2-tuple shape', ([1, 1], [1, 1]), (2, 2)),
      ('rank 2 std gamma, 2-tuple shape', (np.ones((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted concentration', (1, np.ones(3)), (2, 2)),
      ('broadcasted rate', (np.ones(3), 1), ()),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_shape(distr_params, dict(), sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist = self.distrax_cls(
          concentration=jnp.ones((), dtype), rate=jnp.ones((), dtype))
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std gamma, no shape', (11, 3), ()),
      ('1d std gamma, int shape', (11, 3), 1),
      ('1d std gamma, 1-tuple shape', (11, 3), (1,)),
      ('1d std gamma, 2-tuple shape', (1, 1), (2, 2)),
      ('2d std gamma, no shape', (np.ones(2), np.ones(2)), ()),
      ('2d std gamma, int shape', ([1, 1], [1, 1]), 1),
      ('2d std gamma, 1-tuple shape', (np.ones(2), np.ones(2)), (1,)),
      ('2d std gamma, 2-tuple shape', ([1, 1], [1, 1]), (2, 2)),
      ('rank 2 std gamma, 2-tuple shape', (np.ones((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted concentration', (1, np.ones(3)), (2, 2)),
      ('broadcasted rate', (np.ones(3), 1), ()),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_and_log_prob(
        dist_args=distr_params,
        dist_kwargs=dict(),
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (3.1, 1), 1),
      ('1d dist, 2d value', (0.5, 0.1), np.array([1, 2])),
      ('2d dist, 1d value', (0.5 + np.zeros(2), 0.3 * np.ones(2)), 1),
      ('2d broadcasted dist, 1d value', (0.4 + np.zeros(2), 0.8), 1),
      ('2d dist, 2d value', ([0.1, 0.5], 0.9 * np.ones(2)), np.array([1, 2])),
      ('1d dist, 1d value, edge case', (2.1, 1), 200),
  )
  def test_method_with_value(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    for method in ['log_prob', 'prob', 'cdf', 'log_cdf']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=distr_params,
            call_args=(value,),
            assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('no broadcast', ([0.1, 1.3, 0.5], [0.5, 1.3, 1.5])),
      ('broadcasted concentration', (0.5, [0.5, 1.3, 1.5])),
      ('broadcasted rate', ([0.1, 1.3, 0.5], 0.8)),
  )
  def test_method(self, distr_params):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    for method in ['entropy', 'mean', 'variance', 'stddev', 'mode']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=distr_params,
            assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax')
  )
  def test_with_two_distributions(self, function_string, mode_string):
    rtol = 1e-3
    atol = 1e-4
    rng = np.random.default_rng(42)
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'concentration': np.abs(
                rng.normal(size=(4, 1, 2))).astype(np.float32),
            'rate': np.array(
                [[0.8, 0.2], [0.1, 1.2], [1.4, 3.1]], dtype=np.float32),
        },
        dist2_kwargs={
            'concentration': np.abs(rng.normal(size=(3, 2))).astype(np.float32),
            'rate': 0.1 + rng.uniform(size=(4, 1, 2)).astype(np.float32),
        },
        assertion_fn=lambda x, y: np.testing.assert_allclose(x, y, rtol, atol))

  def test_jitable(self):
    super()._test_jittable(
        (0.1, 1.5), assertion_fn=self.assertion_fn(rtol=2e-2))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    concentration = jnp.array(np.abs(np.random.randn(3, 4, 5)))
    rate = jnp.array(np.abs(np.random.randn(3, 4, 5)))
    dist = self.distrax_cls(concentration, rate)
    self.assertion_fn(rtol=2e-2)(
        dist[slice_].concentration, concentration[slice_])
    self.assertion_fn(rtol=2e-2)(dist[slice_].rate, rate[slice_])

  def test_slice_different_parameterization(self):
    concentration = jnp.array(np.abs(np.random.randn(3, 4, 5)))
    rate = jnp.array(np.abs(np.random.randn(4, 5)))
    dist = self.distrax_cls(concentration, rate)
    self.assertion_fn(rtol=2e-2)(dist[0].concentration, concentration[0])
    self.assertion_fn(rtol=2e-2)(dist[0].rate, rate)  # Not slicing rate.


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `uniform.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import uniform
import jax
from jax import config as jax_config
import jax.numpy as jnp


def setUpModule():
  jax_config.update('jax_enable_x64', True)


class UniformFloat64Test(chex.TestCase):

  def _assert_dtypes(self, dist, dtype):
    """Asserts dist methods' outputs' datatypes."""
    # Sanity check to make sure float64 is enabled.
    x_64 = jnp.zeros([])
    self.assertEqual(jnp.float64, x_64.dtype)

    key = jax.random.PRNGKey(1729)
    z, log_prob = self.variant(
        lambda: dist.sample_and_log_prob(seed=key, sample_shape=[3]))()
    z2 = self.variant(
        lambda: dist.sample(seed=key, sample_shape=[3]))()
    self.assertEqual(dtype, z.dtype)
    self.assertEqual(dtype, z2.dtype)
    self.assertEqual(dtype, log_prob.dtype)
    self.assertEqual(dtype, self.variant(dist.log_prob)(z).dtype)
    self.assertEqual(dtype, self.variant(dist.prob)(z).dtype)
    self.assertEqual(dtype, self.variant(dist.log_cdf)(z).dtype)
    self.assertEqual(dtype, self.variant(dist.cdf)(z).dtype)
    self.assertEqual(dtype, self.variant(dist.entropy)().dtype)
    self.assertEqual(dtype, self.variant(dist.mean)().dtype)
    self.assertEqual(dtype, self.variant(dist.median)().dtype)
    self.assertEqual(dtype, self.variant(dist.stddev)().dtype)
    self.assertEqual(dtype, self.variant(dist.variance)().dtype)
    self.assertEqual(dtype, dist.low.dtype)
    self.assertEqual(dtype, dist.high.dtype)
    self.assertEqual(dtype, dist.range.dtype)
    self.assertEqual(dtype, dist.dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_dtype(self, dtype):
    dist = uniform.Uniform(low=jnp.zeros([], dtype), high=jnp.ones([], dtype))
    self._assert_dtypes(dist, dtype)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Categorical-uniform distributions."""

from typing import Tuple

import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import distribution
from distrax._src.distributions import mixture_same_family
from distrax._src.distributions import uniform
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp


Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class CategoricalUniform(distribution.Distribution):
  """Mixture Categorical-Uniform distribution.

  Given an interval `[a, b]` and a probability vector `p = [p_1, ..., p_K]`, a
  random variable `x` follows a Categorical-Uniform distribution if its PDF
  is `p(x) = p_k / C` if `(k-1)C <= x - a < kC` for any `k = 1, ..., K`,
  where `C = (b-a) / K`, and `p(x) = 0` otherwise.

  Equivalently, the Categorical-Uniform can be understood as a mixture of
  Uniform distributions, with mixture probabilities `p_k` and Uniform component
  distributions with support in `[a + (k-1)C, a + kC]`.
  """

  def __init__(
      self,
      *,
      low: Numeric,
      high: Numeric,
      logits: Array,
  ) -> None:
    """Initializes a CategoricalUniform distribution.

    Args:
      low: The lowest value of the support, denoted `a` in the class
        docstring. It can also be a batch of values.
      high: The highest value of the support, denoted `b` in the class
        docstring. It can also be a batch of values.
      logits: The unnormalized log-probabilities of the mixture. It must be an
        array of length `K`. Additional leading dimensions, if any, index
        batches.
    """
    super().__init__()
    self._low = conversion.as_float_array(low)
    self._high = conversion.as_float_array(high)
    self._logits = conversion.as_float_array(logits)
    if self._logits.ndim < 1:
      raise ValueError(
          'The parameter `logits` must have at least one dimension.')

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return jax.lax.broadcast_shapes(
        self._low.shape, self._high.shape, self._logits.shape[:-1])

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    quantile = jax.random.uniform(key, (n,) + self.batch_shape)
    return self._inverse_cdf(quantile)

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    return self._get_mixture().log_prob(value)

  def entropy(self) -> Array:
    """See `Distribution.entropy`."""
    # The following holds because the components have non-overlapping domains.
    mixture = self._get_mixture()
    return mixture.mixture_distribution.entropy() + jnp.log(
        (self._high - self._low) / self.num_bins)

  def mean(self) -> Array:
    """Calculates the mean."""
    return self._get_mixture().mean()

  def variance(self) -> Array:
    """Calculates the variance."""
    return self._get_mixture().variance()

  def __getitem__(self, index) -> 'CategoricalUniform':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return CategoricalUniform(
        low=self.low[index], high=self.high[index], logits=self.logits[index])

  def _get_category_limits(self) -> Array:
    """Gets limits for each category."""
    return jnp.linspace(self.low, self.high, self.num_bins + 1, axis=-1)

  def _get_mixture(self) -> mixture_same_family.MixtureSameFamily:
    """Gets a mixture distribution."""
    limits = self._get_category_limits()
    return mixture_same_family.MixtureSameFamily(
        components_distribution=uniform.Uniform(
            low=limits[..., :-1], high=limits[..., 1:]),
        mixture_distribution=categorical.Categorical(logits=self.logits),
    )

  def _inverse_cdf(self, quantile):
    """Inverse cumulative density function."""
    probs = jax.nn.softmax(self.logits, axis=-1)
    cum_probs = jnp.cumsum(probs, axis=-1)
    quantile_limits = jnp.concatenate(
        [jnp.zeros_like(cum_probs[..., :1]), cum_probs], axis=-1)
    limits = self._get_category_limits()
    domain_diff = jnp.diff(limits, axis=-1)
    quantile_diff = jnp.diff(quantile_limits, axis=-1)
    slopes = domain_diff / quantile_diff
    quantile_contributions = jnp.minimum(
        quantile_diff,
        jax.nn.relu(quantile[..., None] - quantile_limits[..., :-1]),
    )
    return self.low + jnp.sum(slopes * quantile_contributions, axis=-1)

  @property
  def low(self) -> Array:
    # Broadcasted version of the argument passed in the initializer.
    return jnp.broadcast_to(self._low, self.batch_shape)

  @property
  def high(self) -> Array:
    # Broadcasted version of the argument passed in the initializer.
    return jnp.broadcast_to(self._high, self.batch_shape)

  @property
  def logits(self) -> Array:
    return jnp.broadcast_to(self._logits, self.batch_shape + (self.num_bins,))

  @property
  def num_bins(self) -> int:
    return self._logits.shape[-1]



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Scalar affine bijector."""

from typing import Optional, Tuple, Union

from distrax._src.bijectors import bijector as base
import jax
import jax.numpy as jnp


Array = base.Array
Numeric = Union[Array, float]


class ScalarAffine(base.Bijector):
  """An affine bijector that acts elementwise.

  The bijector is defined as follows:

  - Forward: `y = scale * x + shift`
  - Forward Jacobian determinant: `log|det J(x)| = log|scale|`
  - Inverse: `x = (y - shift) / scale`
  - Inverse Jacobian determinant: `log|det J(y)| = -log|scale|`

  where `scale` and `shift` are the bijector's parameters.
  """

  def __init__(self,
               shift: Numeric,
               scale: Optional[Numeric] = None,
               log_scale: Optional[Numeric] = None):
    """Initializes a ScalarAffine bijector.

    Args:
      shift: the bijector's shift parameter. Can also be batched.
      scale: the bijector's scale parameter. Can also be batched. NOTE: `scale`
        must be non-zero, otherwise the bijector is not invertible. It is the
        user's responsibility to make sure `scale` is non-zero; the class will
        make no attempt to verify this.
      log_scale: the log of the scale parameter. Can also be batched. If
        specified, the bijector's scale is set equal to `exp(log_scale)`. Unlike
        `scale`, `log_scale` is an unconstrained parameter. NOTE: either `scale`
        or `log_scale` can be specified, but not both. If neither is specified,
        the bijector's scale will default to 1.

    Raises:
      ValueError: if both `scale` and `log_scale` are not None.
    """
    super().__init__(event_ndims_in=0, is_constant_jacobian=True)
    self._shift = shift
    if scale is None and log_scale is None:
      self._scale = 1.
      self._inv_scale = 1.
      self._log_scale = 0.
    elif log_scale is None:
      self._scale = scale
      self._inv_scale = 1. / scale
      self._log_scale = jnp.log(jnp.abs(scale))
    elif scale is None:
      self._scale = jnp.exp(log_scale)
      self._inv_scale = jnp.exp(jnp.negative(log_scale))
      self._log_scale = log_scale
    else:
      raise ValueError(
          'Only one of `scale` and `log_scale` can be specified, not both.')
    self._batch_shape = jax.lax.broadcast_shapes(
        jnp.shape(self._shift), jnp.shape(self._scale))

  @property
  def shift(self) -> Numeric:
    """The bijector's shift."""
    return self._shift

  @property
  def log_scale(self) -> Numeric:
    """The log of the bijector's scale."""
    return self._log_scale

  @property
  def scale(self) -> Numeric:
    """The bijector's scale."""
    assert self._scale is not None  # By construction.
    return self._scale

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    batch_shape = jax.lax.broadcast_shapes(self._batch_shape, x.shape)
    batched_scale = jnp.broadcast_to(self._scale, batch_shape)
    batched_shift = jnp.broadcast_to(self._shift, batch_shape)
    return batched_scale * x + batched_shift

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    batch_shape = jax.lax.broadcast_shapes(self._batch_shape, x.shape)
    return jnp.broadcast_to(self._log_scale, batch_shape)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self.forward(x), self.forward_log_det_jacobian(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    batch_shape = jax.lax.broadcast_shapes(self._batch_shape, y.shape)
    batched_inv_scale = jnp.broadcast_to(self._inv_scale, batch_shape)
    batched_shift = jnp.broadcast_to(self._shift, batch_shape)
    return batched_inv_scale * (y - batched_shift)

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    batch_shape = jax.lax.broadcast_shapes(self._batch_shape, y.shape)
    return jnp.broadcast_to(jnp.negative(self._log_scale), batch_shape)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    return self.inverse(y), self.inverse_log_det_jacobian(y)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is ScalarAffine:  # pylint: disable=unidiomatic-typecheck
      return all((
          self.shift is other.shift,
          self.scale is other.scale,
          self.log_scale is other.log_scale,
      ))
    else:
      return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `importance_sampling.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import categorical
from distrax._src.utils import importance_sampling
import jax
import jax.numpy as jnp
import numpy as np


class ImportanceSamplingTest(parameterized.TestCase):

  @chex.all_variants(with_pmap=False)
  def test_importance_sampling_ratios_on_policy(self):
    key = jax.random.PRNGKey(42)
    probs = jnp.array([0.4, 0.2, 0.1, 0.3])
    dist = categorical.Categorical(probs=probs)
    event = dist.sample(seed=key, sample_shape=())

    ratios_fn = self.variant(
        importance_sampling.importance_sampling_ratios)
    rhos = ratios_fn(target_dist=dist, sampling_dist=dist, event=event)

    expected_rhos = jnp.ones_like(rhos)
    np.testing.assert_array_almost_equal(rhos, expected_rhos)

  @chex.all_variants(with_pmap=False)
  def test_importance_sampling_ratios_off_policy(self):
    """Tests for a full batch."""
    pi_logits = np.array([[0.2, 0.8], [0.6, 0.4]], dtype=np.float32)
    pi = categorical.Categorical(logits=pi_logits)
    mu_logits = np.array([[0.8, 0.2], [0.6, 0.4]], dtype=np.float32)
    mu = categorical.Categorical(logits=mu_logits)
    events = np.array([1, 0], dtype=np.int32)

    ratios_fn = self.variant(
        importance_sampling.importance_sampling_ratios)
    rhos = ratios_fn(pi, mu, events)

    expected_rhos = np.array(
        [pi.probs[0][1] / mu.probs[0][1], pi.probs[1][0] / mu.probs[1][0]],
        dtype=np.float32)
    np.testing.assert_allclose(expected_rhos, rhos, atol=1e-4)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Gumbel distribution."""

import math
from typing import Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Gumbel(distribution.Distribution):
  """Gumbel distribution with location `loc` and `scale` parameters."""

  equiv_tfp_cls = tfd.Gumbel

  def __init__(self, loc: Numeric, scale: Numeric):
    """Initializes a Gumbel distribution.

    Args:
      loc: Mean of the distribution.
      scale: Spread of the distribution.
    """
    super().__init__()
    self._loc = conversion.as_float_array(loc)
    self._scale = conversion.as_float_array(scale)
    self._batch_shape = jax.lax.broadcast_shapes(
        self._loc.shape, self._scale.shape)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._batch_shape

  @property
  def loc(self) -> Array:
    """Mean of the distribution."""
    return jnp.broadcast_to(self._loc, self.batch_shape)

  @property
  def scale(self) -> Array:
    """Scale of the distribution."""
    return jnp.broadcast_to(self._scale, self.batch_shape)

  def _standardize(self, value: Array) -> Array:
    """Standardizes the input `value` in location and scale."""
    return (value - self._loc) / self._scale

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    z = self._standardize(value)
    return -(z + jnp.exp(-z)) - jnp.log(self._scale)

  def _sample_from_std_gumbel(self, key: PRNGKey, n: int) -> Array:
    out_shape = (n,) + self.batch_shape
    dtype = jnp.result_type(self._loc, self._scale)
    return jax.random.gumbel(key, shape=out_shape, dtype=dtype)

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    rnd = self._sample_from_std_gumbel(key, n)
    return self._scale * rnd + self._loc

  def _sample_n_and_log_prob(self, key: PRNGKey, n: int) -> Tuple[Array, Array]:
    """See `Distribution._sample_n_and_log_prob`."""
    rnd = self._sample_from_std_gumbel(key, n)
    samples = self._scale * rnd + self._loc
    log_prob = -(rnd + jnp.exp(-rnd)) - jnp.log(self._scale)
    return samples, log_prob

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in nats)."""
    return jnp.log(self._scale) + 1. + jnp.euler_gamma

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    z = self._standardize(value)
    return -jnp.exp(-z)

  def mean(self) -> Array:
    """Calculates the mean."""
    return self._loc + self._scale * jnp.euler_gamma

  def stddev(self) -> Array:
    """Calculates the standard deviation."""
    return self._scale * jnp.ones_like(self._loc) * jnp.pi / math.sqrt(6.)

  def variance(self) -> Array:
    """Calculates the variance."""
    return jnp.square(self._scale * jnp.ones_like(self._loc) * jnp.pi) / 6.

  def mode(self) -> Array:
    """Calculates the mode."""
    return self.loc

  def median(self) -> Array:
    """Calculates the median."""
    return self._loc - self._scale * math.log(math.log(2.))

  def __getitem__(self, index) -> 'Gumbel':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Gumbel(loc=self.loc[index], scale=self.scale[index])


def _kl_divergence_gumbel_gumbel(
    dist1: Union[Gumbel, tfd.Gumbel],
    dist2: Union[Gumbel, tfd.Gumbel],
    *unused_args, **unused_kwargs,
    ) -> Array:
  """Batched KL divergence KL(dist1 || dist2) between two Gumbel distributions.

  Args:
    dist1: A Gumbel distribution.
    dist2: A Gumbel distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  return (jnp.log(dist2.scale) - jnp.log(dist1.scale) + jnp.euler_gamma *
          (dist1.scale / dist2.scale - 1.) +
          jnp.expm1((dist2.loc - dist1.loc) / dist2.scale +
                    jax.lax.lgamma(dist1.scale / dist2.scale + 1.)) +
          (dist1.loc - dist2.loc) / dist2.scale)


# Register the KL functions with TFP.
tfd.RegisterKL(Gumbel, Gumbel)(_kl_divergence_gumbel_gumbel)
tfd.RegisterKL(Gumbel, Gumbel.equiv_tfp_cls)(_kl_divergence_gumbel_gumbel)
tfd.RegisterKL(Gumbel.equiv_tfp_cls, Gumbel)(_kl_divergence_gumbel_gumbel)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Laplace distribution."""

import math
from typing import Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


def _log_cdf_laplace(norm_value: EventT) -> Array:
  """Log CDF of a standardized Laplace distribution."""
  lower_value = norm_value - math.log(2.0)
  exp_neg_norm_value = jnp.exp(-jnp.abs(norm_value))
  upper_value = jnp.log1p(-0.5 * exp_neg_norm_value)
  return jnp.where(jnp.less_equal(norm_value, 0.0), lower_value, upper_value)  # pylint: disable=not-callable


class Laplace(distribution.Distribution):
  """Laplace distribution with location `loc` and `scale` parameters."""

  equiv_tfp_cls = tfd.Laplace

  def __init__(self, loc: Numeric, scale: Numeric):
    """Initializes a Laplace distribution.

    Args:
      loc: Mean of the distribution.
      scale: Spread of the distribution.
    """
    super().__init__()
    self._loc = conversion.as_float_array(loc)
    self._scale = conversion.as_float_array(scale)
    self._batch_shape = jax.lax.broadcast_shapes(
        self._loc.shape, self._scale.shape)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._batch_shape

  @property
  def loc(self) -> Array:
    """Mean of the distribution."""
    return jnp.broadcast_to(self._loc, self.batch_shape)

  @property
  def scale(self) -> Array:
    """Scale of the distribution."""
    return jnp.broadcast_to(self._scale, self.batch_shape)

  def _sample_from_std_laplace(self, key: PRNGKey, n: int) -> Array:
    out_shape = (n,) + self.batch_shape
    dtype = jnp.result_type(self._loc, self._scale)
    return jax.random.laplace(key, shape=out_shape, dtype=dtype)

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    rnd = self._sample_from_std_laplace(key, n)
    return self._loc + self._scale * rnd

  def _sample_n_and_log_prob(self, key: PRNGKey, n: int) -> Tuple[Array, Array]:
    """See `Distribution._sample_n_and_log_prob`."""
    rnd = self._sample_from_std_laplace(key, n)
    samples = self._loc + self._scale * rnd
    log_prob = -jnp.abs(rnd) - math.log(2.) - jnp.log(self._scale)
    return samples, log_prob

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    norm_value = self._standardize(value)
    return -jnp.abs(norm_value) - math.log(2.) - jnp.log(self._scale)

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in nats)."""
    return math.log(2.) + 1. + jnp.log(self.scale)

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    norm_value = self._standardize(value)
    return 0.5 - 0.5 * jnp.sign(norm_value) * jnp.expm1(-jnp.abs(norm_value))

  def _standardize(self, value: Array) -> Array:
    return (value - self._loc) / self._scale

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    norm_value = self._standardize(value)
    return _log_cdf_laplace(norm_value)

  def log_survival_function(self, value: EventT) -> Array:
    """See `Distribution.log_survival_function`."""
    norm_value = self._standardize(value)
    return _log_cdf_laplace(-norm_value)

  def mean(self) -> Array:
    """Calculates the mean."""
    return self.loc

  def stddev(self) -> Array:
    """Calculates the standard deviation."""
    return math.sqrt(2.) * self.scale

  def variance(self) -> Array:
    """Calculates the variance."""
    return 2. * jnp.square(self.scale)

  def mode(self) -> Array:
    """Calculates the mode."""
    return self.mean()

  def median(self) -> Array:
    """Calculates the median."""
    return self.mean()

  def __getitem__(self, index) -> 'Laplace':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Laplace(loc=self.loc[index], scale=self.scale[index])


def _kl_divergence_laplace_laplace(
    dist1: Union[Laplace, tfd.Laplace],
    dist2: Union[Laplace, tfd.Laplace],
    *unused_args, **unused_kwargs,
    ) -> Array:
  """Batched KL divergence KL(dist1 || dist2) between two Laplace distributions.

  Args:
    dist1: A Laplace distribution.
    dist2: A Laplace distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  distance = jnp.abs(dist1.loc - dist2.loc)
  diff_log_scale = jnp.log(dist1.scale) - jnp.log(dist2.scale)
  return (- diff_log_scale +
          distance / dist2.scale - 1. +
          jnp.exp(-distance / dist1.scale + diff_log_scale))


# Register the KL functions with TFP.
tfd.RegisterKL(Laplace, Laplace)(_kl_divergence_laplace_laplace)
tfd.RegisterKL(Laplace, Laplace.equiv_tfp_cls)(_kl_divergence_laplace_laplace)
tfd.RegisterKL(Laplace.equiv_tfp_cls, Laplace)(_kl_divergence_laplace_laplace)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Straight-through gradient sampling distribution."""
from distrax._src.distributions import categorical
from distrax._src.distributions import distribution
import jax


def straight_through_wrapper(  # pylint: disable=invalid-name
    Distribution,
    ) -> distribution.DistributionLike:
  """Wrap a distribution to use straight-through gradient for samples."""

  def sample(self, seed, sample_shape=()):  # pylint: disable=g-doc-args
    """Sampling with straight through biased gradient estimator.

    Sample a value from the distribution, but backpropagate through the
    underlying probability to compute the gradient.

    References:
      [1] Yoshua Bengio, Nicholas Léonard, Aaron Courville, Estimating or
      Propagating Gradients Through Stochastic Neurons for Conditional
      Computation, https://arxiv.org/abs/1308.3432

    Args:
      seed: a random seed.
      sample_shape: the shape of the required sample.

    Returns:
      A sample with straight-through gradient.
    """
    # pylint: disable=protected-access
    obj = Distribution(probs=self._probs, logits=self._logits)
    assert isinstance(obj, categorical.Categorical)
    sample = obj.sample(seed=seed, sample_shape=sample_shape)
    probs = obj.probs
    padded_probs = _pad(probs, sample.shape)

    # Keep sample unchanged, but add gradient through probs.
    sample += padded_probs - jax.lax.stop_gradient(padded_probs)
    return sample

  def _pad(probs, shape):
    """Grow probs to have the same number of dimensions as shape."""
    while len(probs.shape) < len(shape):
      probs = probs[None]
    return probs

  parent_name = Distribution.__name__
  # Return a new object, overriding sample.
  return type('StraighThrough' + parent_name, (Distribution,),
              {'sample': sample})



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Multinomial distribution."""

import functools
import operator

from typing import Any, Tuple, Optional, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
from distrax._src.utils import math
import jax
from jax import lax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Multinomial(distribution.Distribution):
  """Multinomial distribution with parameter `probs`."""

  equiv_tfp_cls = tfd.Multinomial

  def __init__(self,
               total_count: Numeric,
               logits: Optional[Array] = None,
               probs: Optional[Array] = None,
               dtype: Union[jnp.dtype, type[Any]] = int):
    """Initializes a Multinomial distribution.

    Args:
      total_count: The number of trials per sample.
      logits: Logit transform of the probability of each category. Only one
        of `logits` or `probs` can be specified.
      probs: Probability of each category. Only one of `logits` or `probs` can
        be specified.
      dtype:  The type of event samples.
    """
    super().__init__()
    logits = None if logits is None else conversion.as_float_array(logits)
    probs = None if probs is None else conversion.as_float_array(probs)
    if (logits is None) == (probs is None):
      raise ValueError(
          f'One and exactly one of `logits` and `probs` should be `None`, '
          f'but `logits` is {logits} and `probs` is {probs}.')
    if logits is not None and (not logits.shape or logits.shape[-1] < 2):
      raise ValueError(
          f'The last dimension of `logits` must be greater than 1, but '
          f'`logits.shape = {logits.shape}`.')
    if probs is not None and (not probs.shape or probs.shape[-1] < 2):
      raise ValueError(
          f'The last dimension of `probs` must be greater than 1, but '
          f'`probs.shape = {probs.shape}`.')
    if not (jnp.issubdtype(dtype, jnp.integer) or
            jnp.issubdtype(dtype, jnp.floating)):
      raise ValueError(
          f'The dtype of `{self.name}` must be integer or floating-point, '
          f'instead got `{dtype}`.')

    self._total_count = jnp.asarray(total_count, dtype=dtype)
    self._probs = None if probs is None else math.normalize(probs=probs)
    self._logits = None if logits is None else math.normalize(logits=logits)
    self._dtype = dtype

    if self._probs is not None:
      probs_batch_shape = self._probs.shape[:-1]
    else:
      assert self._logits is not None
      probs_batch_shape = self._logits.shape[:-1]
    self._batch_shape = lax.broadcast_shapes(
        probs_batch_shape, self._total_count.shape)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    if self._logits is not None:
      return self._logits.shape[-1:]
    else:
      return self._probs.shape[-1:]

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._batch_shape

  @property
  def total_count(self) -> Array:
    """The number of trials per sample."""
    return jnp.broadcast_to(self._total_count, self.batch_shape)

  @property
  def num_trials(self) -> Array:
    """The number of trials for each event."""
    return self.total_count

  @property
  def logits(self) -> Array:
    """The logits for each event."""
    if self._logits is not None:
      return jnp.broadcast_to(self._logits, self.batch_shape + self.event_shape)
    return jnp.broadcast_to(jnp.log(self._probs),
                            self.batch_shape + self.event_shape)

  @property
  def probs(self) -> Array:
    """The probabilities for each event."""
    if self._probs is not None:
      return jnp.broadcast_to(self._probs, self.batch_shape + self.event_shape)
    return jnp.broadcast_to(jax.nn.softmax(self._logits, axis=-1),
                            self.batch_shape + self.event_shape)

  @property
  def log_of_probs(self) -> Array:
    """The log probabilities for each event."""
    if self._logits is not None:
      # jax.nn.log_softmax was already applied in init to logits.
      return jnp.broadcast_to(self._logits,
                              self.batch_shape + self.event_shape)
    return jnp.broadcast_to(jnp.log(self._probs),
                            self.batch_shape + self.event_shape)

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    total_permutations = lax.lgamma(self._total_count + 1.)
    counts_factorial = lax.lgamma(value + 1.)
    redundant_permutations = jnp.sum(counts_factorial, axis=-1)
    log_combinations = total_permutations - redundant_permutations
    return log_combinations + jnp.sum(
        math.multiply_no_nan(self.log_of_probs, value), axis=-1)

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    num_keys = functools.reduce(operator.mul, self.batch_shape, 1)
    keys = jax.random.split(key, num=num_keys)
    total_count = jnp.reshape(self.total_count, (-1,))
    logits = jnp.reshape(self.logits, (-1,) + self.event_shape)
    sample_fn = jax.vmap(
        self._sample_n_scalar, in_axes=(0, 0, None, 0, None), out_axes=1)
    samples = sample_fn(keys, total_count, n, logits, self._dtype)  # [n, B, K]
    return samples.reshape((n,) + self.batch_shape + self.event_shape)

  @staticmethod
  def _sample_n_scalar(
      key: PRNGKey, total_count: Union[int, Array], n: int, logits: Array,
      dtype: jnp.dtype) -> Array:
    """Sample method for a Multinomial with integer `total_count`."""

    def cond_func(args):
      i, _, _ = args
      return jnp.less(i, total_count)  # pylint: disable=not-callable

    def body_func(args):
      i, key_i, sample_aggregator = args
      key_i, current_key = jax.random.split(key_i)
      sample_i = jax.random.categorical(current_key, logits=logits, shape=(n,))
      one_hot_i = jax.nn.one_hot(sample_i, logits.shape[0]).astype(dtype)
      return i + 1, key_i, sample_aggregator + one_hot_i

    init_aggregator = jnp.zeros((n, logits.shape[0]), dtype=dtype)
    return lax.while_loop(cond_func, body_func, (0, key, init_aggregator))[2]

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in nats)."""
    # The method `_entropy_scalar` does not work when `self.total_count` is an
    # array (instead of a scalar) or when we jit the function, so we default to
    # computing the entropy using an alternative method that uses a lax while
    # loop and does not create intermediate arrays whose shape depends on
    # `self.total_count`.
    entropy_fn = jnp.vectorize(
        self._entropy_scalar_with_lax, signature='(),(k),(k)->()')
    return entropy_fn(self.total_count, self.probs, self.log_of_probs)

  @staticmethod
  def _entropy_scalar(
      total_count: int, probs: Array, log_of_probs: Array
    ) -> Union[jnp.float32, jnp.float64]:
    """Calculates the entropy for a Multinomial with integer `total_count`."""
    # Constant factors in the entropy.
    xi = jnp.arange(total_count + 1, dtype=probs.dtype)
    log_xi_factorial = lax.lgamma(xi + 1)
    log_n_minus_xi_factorial = jnp.flip(log_xi_factorial, axis=-1)
    log_n_factorial = log_xi_factorial[..., -1]
    log_comb_n_xi = (
        log_n_factorial[..., None] - log_xi_factorial
        - log_n_minus_xi_factorial)
    comb_n_xi = jnp.round(jnp.exp(log_comb_n_xi))
    chex.assert_shape(comb_n_xi, (total_count + 1,))

    likelihood1 = math.power_no_nan(probs[..., None], xi)
    likelihood2 = math.power_no_nan(1. - probs[..., None], total_count - xi)
    chex.assert_shape(likelihood1, (probs.shape[-1], total_count + 1,))
    chex.assert_shape(likelihood2, (probs.shape[-1], total_count + 1,))
    likelihood = jnp.sum(likelihood1 * likelihood2, axis=-2)
    chex.assert_shape(likelihood, (total_count + 1,))
    comb_term = jnp.sum(comb_n_xi * log_xi_factorial * likelihood, axis=-1)
    chex.assert_shape(comb_term, ())

    # Probs factors in the entropy.
    n_probs_factor = jnp.sum(
        total_count * math.multiply_no_nan(log_of_probs, probs), axis=-1)

    return - log_n_factorial - n_probs_factor + comb_term

  @staticmethod
  def _entropy_scalar_with_lax(
      total_count: int, probs: Array, log_of_probs: Array
    ) -> Union[jnp.float32, jnp.float64]:
    """Like `_entropy_scalar`, but uses a lax while loop."""

    dtype = probs.dtype
    log_n_factorial = lax.lgamma(jnp.asarray(total_count + 1, dtype=dtype))

    def cond_func(args):
      xi, _ = args
      return jnp.less_equal(xi, total_count)  # pylint: disable=not-callable

    def body_func(args):
      xi, accumulated_sum = args
      xi_float = jnp.asarray(xi, dtype=dtype)
      log_xi_factorial = lax.lgamma(xi_float + 1.)
      log_comb_n_xi = (log_n_factorial - log_xi_factorial
                       - lax.lgamma(total_count - xi_float + 1.))
      comb_n_xi = jnp.round(jnp.exp(log_comb_n_xi))
      likelihood1 = math.power_no_nan(probs, xi)
      likelihood2 = math.power_no_nan(1. - probs, total_count - xi)
      likelihood = likelihood1 * likelihood2
      comb_term = comb_n_xi * log_xi_factorial * likelihood  # [K]
      chex.assert_shape(comb_term, (probs.shape[-1],))
      return xi + 1, accumulated_sum + comb_term

    comb_term = jnp.sum(
        lax.while_loop(cond_func, body_func, (0, jnp.zeros_like(probs)))[1],
        axis=-1)

    n_probs_factor = jnp.sum(
        total_count * math.multiply_no_nan(log_of_probs, probs), axis=-1)

    return - log_n_factorial - n_probs_factor + comb_term

  def mean(self) -> Array:
    """Calculates the mean."""
    return self._total_count[..., None] * self.probs

  def variance(self) -> Array:
    """Calculates the variance."""
    probs = self.probs
    return self._total_count[..., None] * probs * (1. - probs)

  def covariance(self) -> Array:
    """Calculates the covariance."""
    probs = self.probs
    cov_matrix = -self._total_count[..., None, None] * (
        probs[..., None, :] * probs[..., :, None])
    chex.assert_shape(cov_matrix, probs.shape + self.event_shape)
    # Missing diagonal term in the covariance matrix.
    cov_matrix += jnp.vectorize(
        jnp.diag, signature='(k)->(k,k)')(
            self._total_count[..., None] * probs)
    return cov_matrix

  def __getitem__(self, index) -> 'Multinomial':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    total_count = self.total_count[index]
    if self._logits is not None:
      return Multinomial(
          total_count=total_count, logits=self.logits[index], dtype=self._dtype)
    return Multinomial(
        total_count=total_count, probs=self.probs[index], dtype=self._dtype)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `inverse.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import bijector as base_bijector
from distrax._src.bijectors import inverse
from distrax._src.bijectors import scalar_affine
from distrax._src.distributions import normal
from distrax._src.distributions import transformed
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions
tfb = tfp.bijectors


RTOL = 1e-2


def _with_additional_parameters(params, all_named_parameters):
  """Convenience function for appending a cartesian product of parameters."""
  for name, param in params:
    for named_params in all_named_parameters:
      yield (f'{named_params[0]}; {name}',) + named_params[1:] + (param,)


def _with_base_dists(*all_named_parameters):
  """Partial of _with_additional_parameters to specify distrax and TFP base."""
  base_dists = (
      ('tfp_base', tfd.Normal),
      ('distrax_base', normal.Normal),
  )
  return _with_additional_parameters(base_dists, all_named_parameters)


class InverseTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.seed = jax.random.PRNGKey(1234)

  def test_properties(self):
    bijector = inverse.Inverse(tfb.Scale(2))
    assert isinstance(bijector.bijector, base_bijector.Bijector)

  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal', 0, 1),
      ('2d std normal', np.zeros(2), np.ones(2)),
      ('broadcasted loc', 0, np.ones(3)),
      ('broadcasted scale', np.ones(3), 1),
  ))
  def test_event_shape(self, mu, sigma, base_dist):
    base = base_dist(mu, sigma)
    bijector = inverse.Inverse(tfb.Scale(2))
    dist = transformed.Transformed(base, bijector)

    tfp_bijector = tfb.Invert(tfb.Scale(2))
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)

    assert dist.event_shape == tfp_dist.event_shape

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal, no shape', 0, 1, ()),
      ('1d std normal, int shape', 0, 1, 1),
      ('1d std normal, 1-tuple shape', 0, 1, (1,)),
      ('1d std normal, 2-tuple shape', 0, 1, (2, 2)),
      ('2d std normal, no shape', np.zeros(2), np.ones(2), ()),
      ('2d std normal, int shape', [0, 0], [1, 1], 1),
      ('2d std normal, 1-tuple shape', np.zeros(2), np.ones(2), (1,)),
      ('2d std normal, 2-tuple shape', [0, 0], [1, 1], (2, 2)),
      ('rank 2 std normal, 2-tuple shape', np.zeros(
          (3, 2)), np.ones((3, 2)), (2, 2)),
      ('broadcasted loc', 0, np.ones(3), (2, 2)),
      ('broadcasted scale', np.ones(3), 1, ()),
  ))
  def test_sample_shape(self, mu, sigma, sample_shape, base_dist):
    base = base_dist(mu, sigma)
    bijector = inverse.Inverse(tfb.Scale(2))
    dist = transformed.Transformed(base, bijector)
    def sample_fn(seed, sample_shape):
      return dist.sample(seed=seed, sample_shape=sample_shape)
    samples = self.variant(sample_fn, ignore_argnums=(1,), static_argnums=1)(
        self.seed, sample_shape)

    tfp_bijector = tfb.Invert(tfb.Scale(2))
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    tfp_samples = tfp_dist.sample(sample_shape=sample_shape, seed=self.seed)

    chex.assert_equal_shape([samples, tfp_samples])

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d dist, 1d value', 0, 1, 1.),
      ('1d dist, 2d value', 0., 1., np.array([1., 2.])),
      ('2d dist, 1d value', np.zeros(2), np.ones(2), 1.),
      ('2d broadcasted dist, 1d value', np.zeros(2), 1, 1.),
      ('2d dist, 2d value', np.zeros(2), np.ones(2), np.array([1., 2.])),
      ('1d dist, 1d value, edge case', 0, 1, 200.),
  ))
  def test_log_prob(self, mu, sigma, value, base_dist):
    base = base_dist(mu, sigma)
    bijector = inverse.Inverse(tfb.Scale(2))
    dist = transformed.Transformed(base, bijector)
    actual = self.variant(dist.log_prob)(value)

    tfp_bijector = tfb.Invert(tfb.Scale(2))
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    expected = tfp_dist.log_prob(value)
    np.testing.assert_allclose(actual, expected, atol=1e-6)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d dist, 1d value', 0, 1, 1.),
      ('1d dist, 2d value', 0., 1., np.array([1., 2.])),
      ('2d dist, 1d value', np.zeros(2), np.ones(2), 1.),
      ('2d broadcasted dist, 1d value', np.zeros(2), 1, 1.),
      ('2d dist, 2d value', np.zeros(2), np.ones(2), np.array([1., 2.])),
      ('1d dist, 1d value, edge case', 0, 1, 200.),
  ))
  def test_prob(self, mu, sigma, value, base_dist):
    base = base_dist(mu, sigma)
    bijector = inverse.Inverse(tfb.Scale(2))
    dist = transformed.Transformed(base, bijector)
    actual = self.variant(dist.prob)(value)

    tfp_bijector = tfb.Invert(tfb.Scale(2))
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    expected = tfp_dist.prob(value)
    np.testing.assert_allclose(actual, expected, atol=1e-9)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal, no shape', 0, 1, ()),
      ('1d std normal, int shape', 0, 1, 1),
      ('1d std normal, 1-tuple shape', 0, 1, (1,)),
      ('1d std normal, 2-tuple shape', 0, 1, (2, 2)),
      ('2d std normal, no shape', np.zeros(2), np.ones(2), ()),
      ('2d std normal, int shape', [0, 0], [1, 1], 1),
      ('2d std normal, 1-tuple shape', np.zeros(2), np.ones(2), (1,)),
      ('2d std normal, 2-tuple shape', [0, 0], [1, 1], (2, 2)),
      ('rank 2 std normal, 2-tuple shape', np.zeros(
          (3, 2)), np.ones((3, 2)), (2, 2)),
      ('broadcasted loc', 0, np.ones(3), (2, 2)),
      ('broadcasted scale', np.ones(3), 1, ()),
  ))
  def test_sample_and_log_prob(self, mu, sigma, sample_shape, base_dist):
    base = base_dist(mu, sigma)
    bijector = inverse.Inverse(tfb.Scale(2))
    dist = transformed.Transformed(base, bijector)
    def sample_and_log_prob_fn(seed, sample_shape):
      return dist.sample_and_log_prob(seed=seed, sample_shape=sample_shape)
    samples, log_prob = self.variant(
        sample_and_log_prob_fn, ignore_argnums=(1,), static_argnums=(1,))(
            self.seed, sample_shape)
    expected_samples = bijector.forward(
        base.sample(seed=self.seed, sample_shape=sample_shape))

    tfp_bijector = tfb.Invert(tfb.Scale(2))
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)
    tfp_samples = tfp_dist.sample(seed=self.seed, sample_shape=sample_shape)
    tfp_log_prob = tfp_dist.log_prob(samples)

    chex.assert_equal_shape([samples, tfp_samples])
    np.testing.assert_allclose(log_prob, tfp_log_prob, rtol=RTOL)
    np.testing.assert_allclose(samples, expected_samples, rtol=RTOL)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(_with_base_dists(
      ('entropy', 'entropy', 0., 1.),
      ('mean', 'mean', 0, 1),
      ('mean from list params', 'mean', [-1, 1], [1, 2]),
      ('mode', 'mode', 0, 1),
  ))
  def test_method(self, function_string, mu, sigma, base_dist):
    base = base_dist(mu, sigma)
    bijector = inverse.Inverse(tfb.Scale(2))
    dist = transformed.Transformed(base, bijector)

    tfp_bijector = tfb.Invert(tfb.Scale(2))
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)

    np.testing.assert_allclose(
        self.variant(getattr(dist, function_string))(),
        getattr(tfp_dist, function_string)())

  @chex.all_variants
  @parameterized.named_parameters(
      ('int16', np.array([1, 2], dtype=np.int16)),
      ('int32', np.array([1, 2], dtype=np.int32)),
      ('int64', np.array([1, 2], dtype=np.int64)),
  )
  def test_integer_inputs(self, inputs):
    bijector = inverse.Inverse(scalar_affine.ScalarAffine(shift=1.0))
    output, log_det = self.variant(bijector.forward_and_log_det)(inputs)

    expected_out = jnp.array(inputs, dtype=jnp.float32) - 1.0
    expected_log_det = jnp.zeros_like(inputs, dtype=jnp.float32)

    np.testing.assert_array_equal(output, expected_out)
    np.testing.assert_array_equal(log_det, expected_log_det)

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = inverse.Inverse(scalar_affine.ScalarAffine(0, 1))
    x = np.zeros(())
    f(x, bijector)

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Sigmoid bijector."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
import jax
import jax.numpy as jnp


Array = base.Array


class Sigmoid(base.Bijector):
  """A bijector that computes the logistic sigmoid.

  The log-determinant implementation in this bijector is more numerically stable
  than relying on the automatic differentiation approach used by Lambda, so this
  bijector should be preferred over Lambda(jax.nn.sigmoid) where possible. See
  `tfp.bijectors.Sigmoid` for details.

  Note that the underlying implementation of `jax.nn.sigmoid` used by the
  `forward` function of this bijector does not support inputs of integer type.
  To invoke the forward function of this bijector on an argument of integer
  type, it should first be cast explicitly to a floating point type.

  When the absolute value of the input is large, `Sigmoid` becomes close to a
  constant, so that it is not possible to recover the input `x` from the output
  `y` within machine precision. In cases where it is needed to compute both the
  forward mapping and the backward mapping one after the other to recover the
  original input `x`, it is the user's responsibility to simplify the operation
  to avoid numerical issues; this is unlike the `tfp.bijectors.Sigmoid`. One
  example of such case is to use the bijector within a `Transformed`
  distribution and to obtain the log-probability of samples obtained from the
  distribution's `sample` method. For values of the samples for which it is not
  possible to apply the inverse bijector accurately, `log_prob` returns NaN.
  This can be avoided by using `sample_and_log_prob` instead of `sample`
  followed by `log_prob`.
  """

  def __init__(self):
    """Initializes a Sigmoid bijector."""
    super().__init__(event_ndims_in=0)

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    # pylint:disable=invalid-unary-operand-type
    return -_more_stable_softplus(-x) - _more_stable_softplus(x)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return _more_stable_sigmoid(x), self.forward_log_det_jacobian(x)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    x = jnp.log(y) - jnp.log1p(-y)
    return x, -self.forward_log_det_jacobian(x)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    return type(other) is Sigmoid  # pylint: disable=unidiomatic-typecheck


def _more_stable_sigmoid(x: Array) -> Array:
  """Where extremely negatively saturated, approximate sigmoid with exp(x)."""
  return jnp.where(x < -9, jnp.exp(x), jax.nn.sigmoid(x))


def _more_stable_softplus(x: Array) -> Array:
  """Where extremely saturated, approximate softplus with log1p(exp(x))."""
  return jnp.where(x < -9, jnp.log1p(jnp.exp(x)), jax.nn.softplus(x))



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MultivariateNormalDiag distribution."""

from typing import Optional

import chex
from distrax._src.bijectors.diag_linear import DiagLinear
from distrax._src.distributions import distribution
from distrax._src.distributions.mvn_from_bijector import MultivariateNormalFromBijector
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
EventT = distribution.EventT


def _check_parameters(
    loc: Optional[Array], scale_diag: Optional[Array]) -> None:
  """Checks that the `loc` and `scale_diag` parameters are correct."""
  chex.assert_not_both_none(loc, scale_diag)
  if scale_diag is not None and not scale_diag.shape:
    raise ValueError('If provided, argument `scale_diag` must have at least '
                     '1 dimension.')
  if loc is not None and not loc.shape:
    raise ValueError('If provided, argument `loc` must have at least '
                     '1 dimension.')
  if loc is not None and scale_diag is not None and (
      loc.shape[-1] != scale_diag.shape[-1]):
    raise ValueError(f'The last dimension of arguments `loc` and '
                     f'`scale_diag` must coincide, but {loc.shape[-1]} != '
                     f'{scale_diag.shape[-1]}.')


class MultivariateNormalDiag(MultivariateNormalFromBijector):
  """Multivariate normal distribution on `R^k` with diagonal covariance."""

  equiv_tfp_cls = tfd.MultivariateNormalDiag

  def __init__(self,
               loc: Optional[Array] = None,
               scale_diag: Optional[Array] = None):
    """Initializes a MultivariateNormalDiag distribution.

    Args:
      loc: Mean vector of the distribution. Can also be a batch of vectors. If
        not specified, it defaults to zeros. At least one of `loc` and
        `scale_diag` must be specified.
      scale_diag: Vector of standard deviations. Can also be a batch of vectors.
        If not specified, it defaults to ones. At least one of `loc` and
        `scale_diag` must be specified.
    """
    _check_parameters(loc, scale_diag)

    if scale_diag is None:
      loc = conversion.as_float_array(loc)
      scale_diag = jnp.ones(loc.shape[-1], loc.dtype)
    elif loc is None:
      scale_diag = conversion.as_float_array(scale_diag)
      loc = jnp.zeros(scale_diag.shape[-1], scale_diag.dtype)
    else:
      loc = conversion.as_float_array(loc)
      scale_diag = conversion.as_float_array(scale_diag)

    # Add leading dimensions to the paramteters to match the batch shape. This
    # prevents automatic rank promotion.
    broadcasted_shapes = jnp.broadcast_shapes(loc.shape, scale_diag.shape)
    loc = jnp.expand_dims(
        loc, axis=list(range(len(broadcasted_shapes) - loc.ndim)))
    scale_diag = jnp.expand_dims(
        scale_diag, axis=list(range(len(broadcasted_shapes) - scale_diag.ndim)))

    bias = jnp.zeros_like(loc, shape=loc.shape[-1:])
    bias = jnp.expand_dims(
        bias, axis=list(range(len(broadcasted_shapes) - bias.ndim)))
    scale = DiagLinear(scale_diag)
    super().__init__(loc=loc, scale=scale)
    self._scale_diag = scale_diag

  @property
  def scale_diag(self) -> Array:
    """Scale of the distribution."""
    return jnp.broadcast_to(
        self._scale_diag, self.batch_shape + self.event_shape)

  def _standardize(self, value: Array) -> Array:
    return (value - self._loc) / self._scale_diag

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    return jnp.prod(jax.scipy.special.ndtr(self._standardize(value)), axis=-1)

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return jnp.sum(
        jax.scipy.special.log_ndtr(self._standardize(value)), axis=-1)

  def __getitem__(self, index) -> 'MultivariateNormalDiag':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return MultivariateNormalDiag(
        loc=self.loc[index], scale_diag=self.scale_diag[index])



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `bernoulli.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import bernoulli
from distrax._src.utils import equivalence
import jax.numpy as jnp
import numpy as np
from scipy import special as sp_special


class BernoulliTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(bernoulli.Bernoulli)
    self.p = np.asarray([0.2, 0.4, 0.6, 0.8])
    self.logits = sp_special.logit(self.p)

  @parameterized.named_parameters(
      ('0d probs', (), True),
      ('0d logits', (), False),
      ('1d probs', (4,), True),
      ('1d logits', (4,), False),
      ('2d probs', (3, 4), True),
      ('2d logits', (3, 4), False),
  )
  def test_properties(self, shape, from_probs):
    rng = np.random.default_rng(42)
    probs = rng.uniform(size=shape)
    logits = sp_special.logit(probs)
    dist_kwargs = {'probs': probs} if from_probs else {'logits': logits}
    dist = self.distrax_cls(**dist_kwargs)
    self.assertion_fn(rtol=1e-3)(dist.logits, logits)
    self.assertion_fn(rtol=1e-3)(dist.probs, probs)
    self.assertEqual(dist.event_shape, ())
    self.assertEqual(dist.batch_shape, shape)

  @parameterized.named_parameters(
      ('probs and logits', {'logits': [0.1, -0.2], 'probs': [0.5, 0.4]}),
      ('both probs and logits are None', {'logits': None, 'probs': None}),
      ('complex64 dtype', {'logits': [0.1, -0.2], 'dtype': jnp.complex64}),
      ('complex128 dtype', {'logits': [0.1, -0.2], 'dtype': jnp.complex128}),
  )
  def test_raises_on_invalid_inputs(self, dist_params):
    with self.assertRaises(ValueError):
      self.distrax_cls(**dist_params)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape', {'logits': [0.0, 1.0, -0.5]}, ()),
      ('1d probs, no shape', {'probs': [0.1, 0.5, 0.3]}, ()),
      ('1d logits, int shape', {'logits': [0.0, 1.0, -0.5]}, 1),
      ('1d probs, int shape', {'probs': [0.1, 0.5, 0.3]}, 1),
      ('1d logits, 1-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (1,)),
      ('1d probs, 1-tuple shape', {'probs': [0.1, 0.5, 0.3]}, (1,)),
      ('1d logits, 2-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (5, 4)),
      ('1d probs, 2-tuple shape', {'probs': [0.1, 0.5, 0.3]}, (5, 4)),
      ('2d logits, no shape', {'logits': [[0.0, 1.0, -0.5],
                                          [-0.1, 0.3, 0.0]]}, ()),
      ('2d probs, no shape', {'probs': [[0.1, 0.4, 0.5],
                                        [0.5, 0.25, 0.25]]}, ()),
      ('2d logits, int shape', {'logits': [[0.0, 1.0, -0.5],
                                           [-0.1, 0.3, 0.0]]}, 4),
      ('2d probs, int shape', {'probs': [[0.1, 0.4, 0.5],
                                         [0.5, 0.25, 0.25]]}, 4),
      ('2d logits, 1-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 0.0]]}, (5,)),
      ('2d probs, 1-tuple shape', {'probs': [[0.1, 0.4, 0.5],
                                             [0.5, 0.25, 0.25]]}, (5,)),
      ('2d logits, 2-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 0.0]]}, (5, 4)),
      ('2d probs, 2-tuple shape', {'probs': [[0.1, 0.4, 0.5],
                                             [0.5, 0.25, 0.25]]}, (5, 4)),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    super()._test_sample_shape(
        dist_args=(),
        dist_kwargs=distr_params,
        sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('sample, from probs', 'sample', True),
      ('sample, from logits', 'sample', False),
      ('sample_and_log_prob, from probs', 'sample_and_log_prob', True),
      ('sample_and_log_prob, from logits', 'sample_and_log_prob', False),
  )
  def test_sample_values(self, method, from_probs):
    probs = np.array([0., 0.2, 0.5, 0.8, 1.])  # Includes edge cases (0 and 1).
    logits = sp_special.logit(probs)
    n_samples = 100000
    dist_kwargs = {'probs': probs} if from_probs else {'logits': logits}
    dist = self.distrax_cls(**dist_kwargs)
    sample_fn = self.variant(
        lambda key: getattr(dist, method)(seed=key, sample_shape=n_samples))
    samples = sample_fn(self.key)
    samples = samples[0] if method == 'sample_and_log_prob' else samples
    self.assertEqual(samples.shape, (n_samples,) + probs.shape)
    self.assertTrue(np.all(np.logical_or(samples == 0, samples == 1)))
    self.assertion_fn(rtol=0.1)(np.mean(samples, axis=0), probs)
    self.assertion_fn(atol=2e-3)(np.std(samples, axis=0), dist.stddev())

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape', {'logits': [0.0, 1.0, -0.5]}, ()),
      ('1d probs, no shape', {'probs': [0.1, 0.5, 0.3]}, ()),
      ('1d logits, int shape', {'logits': [0.0, 1.0, -0.5]}, 1),
      ('1d probs, int shape', {'probs': [0.1, 0.5, 0.3]}, 1),
      ('1d logits, 1-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (1,)),
      ('1d probs, 1-tuple shape', {'probs': [0.1, 0.5, 0.3]}, (1,)),
      ('1d logits, 2-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (5, 4)),
      ('1d probs, 2-tuple shape', {'probs': [0.1, 0.5, 0.3]}, (5, 4)),
      ('2d logits, no shape', {'logits': [[0.0, 1.0, -0.5],
                                          [-0.1, 0.3, 0.0]]}, ()),
      ('2d probs, no shape', {'probs': [[0.1, 0.4, 0.5],
                                        [0.5, 0.25, 0.25]]}, ()),
      ('2d logits, int shape', {'logits': [[0.0, 1.0, -0.5],
                                           [-0.1, 0.3, 0.0]]}, 4),
      ('2d probs, int shape', {'probs': [[0.1, 0.4, 0.5],
                                         [0.5, 0.25, 0.25]]}, 4),
      ('2d logits, 1-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 0.0]]}, (5,)),
      ('2d probs, 1-tuple shape', {'probs': [[0.1, 0.4, 0.5],
                                             [0.5, 0.25, 0.25]]}, (5,)),
      ('2d logits, 2-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 0.0]]}, (5, 4)),
      ('2d probs, 2-tuple shape', {'probs': [[0.1, 0.4, 0.5],
                                             [0.5, 0.25, 0.25]]}, (5, 4)),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    super()._test_sample_and_log_prob(
        dist_args=(),
        dist_kwargs=distr_params,
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('sample, bool', 'sample', jnp.bool_),
      ('sample, uint16', 'sample', jnp.uint16),
      ('sample, uint32', 'sample', jnp.uint32),
      ('sample, int16', 'sample', jnp.int16),
      ('sample, int32', 'sample', jnp.int32),
      ('sample, float16', 'sample', jnp.float16),
      ('sample, float32', 'sample', jnp.float32),
      ('sample_and_log_prob, bool', 'sample_and_log_prob', jnp.bool_),
      ('sample_and_log_prob, uint16', 'sample_and_log_prob', jnp.uint16),
      ('sample_and_log_prob, uint32', 'sample_and_log_prob', jnp.uint32),
      ('sample_and_log_prob, int16', 'sample_and_log_prob', jnp.int16),
      ('sample_and_log_prob, int32', 'sample_and_log_prob', jnp.int32),
      ('sample_and_log_prob, float16', 'sample_and_log_prob', jnp.float16),
      ('sample_and_log_prob, float32', 'sample_and_log_prob', jnp.float32),
  )
  def test_sample_dtype(self, method, dtype):
    dist_params = {'logits': self.logits, 'dtype': dtype}
    dist = self.distrax_cls(**dist_params)
    samples = self.variant(getattr(dist, method))(seed=self.key)
    samples = samples[0] if method == 'sample_and_log_prob' else samples
    self.assertEqual(samples.dtype, dist.dtype)
    self.assertEqual(samples.dtype, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, int value', {'logits': [0.0, 0.5, -0.5]}, 1),
      ('1d probs, int value', {'probs': [0.3, 0.2, 0.5]}, 1),
      ('1d logits, 1d value', {'logits': [0.0, 0.5, -0.5]}, [1, 0, 1]),
      ('1d probs, 1d value', {'probs': [0.3, 0.2, 0.5]}, [1, 0, 1]),
      ('1d logits, 2d value', {'logits': [0.0, 0.5, -0.5]},
       [[1, 0, 0], [0, 1, 0]]),
      ('1d probs, 2d value', {'probs': [0.3, 0.2, 0.5]},
       [[1, 0, 0], [0, 1, 0]]),
      ('2d logits, 1d value', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]},
       [1, 0, 1]),
      ('2d probs, 1d value', {'probs': [[0.1, 0.5, 0.4], [0.3, 0.3, 0.4]]},
       [1, 0, 1]),
      ('2d logits, 2d value', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]},
       [[1, 0, 0], [1, 1, 0]]),
      ('2d probs, 2d value', {'probs': [[0.1, 0.5, 0.4], [0.3, 0.3, 0.4]]},
       [[1, 0, 0], [1, 1, 0]]),
      ('edge cases with logits', {'logits': [-np.inf, -np.inf, np.inf, np.inf]},
       [0, 1, 0, 1]),
      ('edge cases with probs', {'probs': [0.0, 0.0, 1.0, 1.0]}, [0, 1, 0, 1]),
  )
  def test_method_with_value(self, distr_params, value):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    value = jnp.asarray(value)
    for method in ['prob', 'log_prob', 'cdf', 'log_cdf',
                   'survival_function', 'log_survival_function']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_kwargs=distr_params,
            call_args=(value,),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('from logits', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]}),
      ('from probs', {'probs': [[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]]}),
  )
  def test_method(self, distr_params):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    for method in ['entropy', 'mode', 'mean', 'variance', 'stddev']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_kwargs=distr_params,
            call_args=(),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'),
  )
  def test_with_two_distributions(self, function_string, mode_string):
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'probs': jnp.asarray([[0.1, 0.5, 0.4], [0.2, 0.4, 0.8]])},
        dist2_kwargs={'logits': jnp.asarray([0.0, -0.1, 0.1]),},
        assertion_fn=self.assertion_fn(rtol=1e-2))

  def test_jittable(self):
    super()._test_jittable(
        (np.array([0., 4., -1., 4.]),),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @parameterized.named_parameters(
      ('single element, from probs', 2, True),
      ('single element, from logits', 2, False),
      ('range, from probs', slice(-1), True),
      ('range, from logits', slice(-1), False),
      ('range_2, from probs', (slice(None), slice(-1)), True),
      ('range_2, from logits', (slice(None), slice(-1)), False),
      ('ellipsis, from probs', (Ellipsis, -1), True),
      ('ellipsis, from logits', (Ellipsis, -1), False),
  )
  def test_slice(self, slice_, from_probs):
    rng = np.random.default_rng(42)
    probs = rng.uniform(size=(3, 4, 5))
    logits = sp_special.logit(probs)
    dist_kwargs = {'probs': probs} if from_probs else {'logits': logits}
    dist = self.distrax_cls(**dist_kwargs)
    self.assertion_fn(rtol=1e-3)(dist[slice_].logits, logits[slice_])
    self.assertion_fn(rtol=1e-3)(dist[slice_].probs, probs[slice_])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper to adapt a TFP distribution."""

from typing import Tuple

import chex
from distrax._src.distributions import distribution
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
PRNGKey = chex.PRNGKey
DistributionT = distribution.DistributionT
EventT = distribution.EventT


def distribution_from_tfp(tfp_distribution: tfd.Distribution) -> DistributionT:
  """Create a Distrax distribution from a TFP distribution.

  Given a TFP distribution `tfp_distribution`, this method returns a
  distribution of a class that inherits from the class of `tfp_distribution`.
  The returned distribution behaves almost identically as the TFP distribution,
  except the common methods (`sample`, `variance`, etc.) are overwritten to
  return `jnp.ndarrays`. Moreover, the wrapped distribution also implements
  Distrax methods inherited from `Distribution`, such as `sample_and_log_prob`.

  Args:
    tfp_distribution: A TFP distribution.

  Returns:
    The wrapped distribution.
  """

  class DistributionFromTFP(
      distribution.Distribution, tfp_distribution.__class__):
    """Class to wrap a TFP distribution.

    The wrapped class dynamically inherits from `tfp_distribution`, so that
    computations involving the KL remain valid.
    """

    def __init__(self):
      pass

    def __getattr__(self, name: str):
      return getattr(tfp_distribution, name)

    def sample(self, *a, **k):  # pylint: disable=useless-super-delegation
      """See `Distribution.sample`."""
      return super().sample(*a, **k)

    def _sample_n(self, key: PRNGKey, n: int):
      """See `Distribution._sample_n`."""
      return jnp.asarray(
          tfp_distribution.sample(seed=key, sample_shape=(n,)),
          dtype=tfp_distribution.dtype)

    def log_prob(self, value: EventT) -> Array:
      """See `Distribution.log_prob`."""
      return jnp.asarray(tfp_distribution.log_prob(value))

    def prob(self, value: EventT) -> Array:
      """See `Distribution.prob`."""
      return jnp.asarray(tfp_distribution.prob(value))

    @property
    def event_shape(self) -> Tuple[int, ...]:
      """See `Distribution.event_shape`."""
      return tuple(tfp_distribution.event_shape)

    @property
    def batch_shape(self) -> Tuple[int, ...]:
      """See `Distribution.batch_shape`."""
      return tuple(tfp_distribution.batch_shape)

    @property
    def name(self) -> str:
      """See `Distribution.name`."""
      return tfp_distribution.name

    @property
    def dtype(self) -> jnp.dtype:
      """See `Distribution.dtype`."""
      return tfp_distribution.dtype

    def kl_divergence(self, other_dist, *args, **kwargs) -> Array:
      """See `Distribution.kl_divergence`."""
      return jnp.asarray(
          tfd.kullback_leibler.kl_divergence(self, other_dist, *args, **kwargs))

    def entropy(self) -> Array:
      """See `Distribution.entropy`."""
      return jnp.asarray(tfp_distribution.entropy())

    def log_cdf(self, value: EventT) -> Array:
      """See `Distribution.log_cdf`."""
      return jnp.asarray(tfp_distribution.log_cdf(value))

    def cdf(self, value: EventT) -> Array:
      """See `Distribution.cdf`."""
      return jnp.asarray(tfp_distribution.cdf(value))

    def mean(self) -> Array:
      """See `Distribution.mean`."""
      return jnp.asarray(tfp_distribution.mean())

    def median(self) -> Array:
      """See `Distribution.median`."""
      return jnp.asarray(tfp_distribution.median())

    def variance(self) -> Array:
      """See `Distribution.variance`."""
      return jnp.asarray(tfp_distribution.variance())

    def stddev(self) -> Array:
      """See `Distribution.stddev`."""
      return jnp.asarray(tfp_distribution.stddev())

    def mode(self) -> Array:
      """See `Distribution.mode`."""
      return jnp.asarray(tfp_distribution.mode())

    def __getitem__(self, index) -> DistributionT:
      """See `Distribution.__getitem__`."""
      return distribution_from_tfp(tfp_distribution[index])

  return DistributionFromTFP()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `log_stddev_normal.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import log_stddev_normal as lsn
from distrax._src.distributions import normal
import jax
import jax.experimental
import jax.numpy as jnp
import mock
import numpy as np
from tensorflow_probability.substrates import jax as tfp


kl_module = tfp.distributions.kullback_leibler


class LogStddevNormalTest(parameterized.TestCase):

  @parameterized.parameters(
      (np.zeros((4,)), np.zeros((4,)), np.zeros((4,))),
      (np.zeros(()), np.zeros((4,)), np.zeros((4,))),
      (np.zeros((4,)), np.zeros(()), np.zeros((4,))),
  )
  def test_log_scale_property(self, mean, log_stddev, expected):
    dist = lsn.LogStddevNormal(mean, log_stddev)
    self.assertEqual(dist.log_scale.shape, expected.shape)
    np.testing.assert_allclose(dist.log_scale, expected, atol=1e-4)

  @parameterized.parameters(
      (0.0, 1.0),
      (4.0, 10.0),
  )
  def test_sampling_scalar(self, mean, stddev):
    log_stddev = np.log(stddev)
    dist = lsn.LogStddevNormal(mean, log_stddev)

    num_samples = 1000000
    prng_key = jax.random.PRNGKey(1331)
    samples = dist.sample(seed=prng_key, sample_shape=num_samples)
    chex.assert_shape(samples, (num_samples,))
    np.testing.assert_allclose(jnp.mean(samples), mean, atol=4e-2)
    np.testing.assert_allclose(jnp.std(samples), stddev, atol=4e-2)

  @parameterized.parameters(
      ([3, 4], [1.5, 2.5]),
      ([0, 1, 0, 1, 10], [0.1, 0.5, 1.0, 5.0, 10.0]))
  def test_sampling_vector(self, mean, stddev):
    mean = np.array(mean)
    log_stddev = np.log(stddev)
    self.assertEqual(mean.shape, log_stddev.shape)
    dist = lsn.LogStddevNormal(mean, log_stddev)

    num_samples = 1000000
    prng_key = jax.random.PRNGKey(1331)
    samples = dist.sample(seed=prng_key, sample_shape=num_samples)
    chex.assert_shape(samples, (num_samples,) + mean.shape)
    np.testing.assert_allclose(jnp.mean(samples, axis=0), mean, atol=4e-2)
    np.testing.assert_allclose(jnp.std(samples, axis=0), stddev, atol=4e-2)

  def test_sampling_batched(self):
    means = np.array([[3.0, 4.0], [-5, 48.0], [58, 64.0]])
    stddevs = np.array([[1, 2], [2, 4], [4, 8]])
    log_stddevs = np.log(stddevs)
    dist = lsn.LogStddevNormal(means, log_stddevs)

    num_samples = 1000000
    prng_key = jax.random.PRNGKey(1331)
    samples = dist.sample(seed=prng_key, sample_shape=num_samples)
    # output shape is [num_samples] + means.shape
    chex.assert_shape(samples, (num_samples, 3, 2))
    np.testing.assert_allclose(jnp.mean(samples, axis=0), means, atol=4e-2)
    np.testing.assert_allclose(jnp.std(samples, axis=0), stddevs, atol=4e-2)

  def test_sampling_batched_custom_dim(self):
    means = np.array([[3.0, 4.0], [-5, 48.0], [58, 64.0]])
    stddevs = np.array([[1, 2], [2, 4], [4, 8]])
    log_stddevs = np.log(stddevs)
    dist = lsn.LogStddevNormal(means, log_stddevs)

    num_samples = 1000000
    prng_key = jax.random.PRNGKey(1331)
    samples = dist.sample(seed=prng_key, sample_shape=num_samples)
    chex.assert_shape(samples, (num_samples, 3, 2))
    np.testing.assert_allclose(jnp.mean(samples, axis=0), means, atol=4e-2)
    np.testing.assert_allclose(jnp.std(samples, axis=0), stddevs, atol=4e-2)

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist = lsn.LogStddevNormal(
          loc=jnp.zeros((), dtype), log_scale=jnp.zeros((), dtype))
      samples = self.variant(dist.sample)(seed=jax.random.PRNGKey(0))
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  def test_kl_versus_normal(self):
    loc, scale = jnp.array([2.0]), jnp.array([2.0])
    log_scale = jnp.log(scale)
    lsn_prior = lsn.LogStddevNormal(jnp.array([0.0]), jnp.array([0.0]))
    n_prior = normal.Normal(jnp.array([0.0]), jnp.array([1.0]))
    lsn_dist = lsn.LogStddevNormal(loc, log_scale)
    n_dist = normal.Normal(loc, scale)

    kl1 = tfp.distributions.kl_divergence(lsn_dist, lsn_prior)
    kl2 = tfp.distributions.kl_divergence(n_dist, lsn_prior)
    kl3 = tfp.distributions.kl_divergence(n_dist, n_prior)
    np.testing.assert_allclose(kl2, kl1)
    np.testing.assert_allclose(kl3, kl2)
    np.testing.assert_allclose(kl1, kl3)

  # pylint:disable=protected-access
  def test_custom_kl_registered(self):
    # Check that our custom KL is registered inside the TFP dispatch table.
    dist_pair = (lsn.LogStddevNormal, lsn.LogStddevNormal)
    self.assertEqual(kl_module._DIVERGENCES[dist_pair],
                     lsn._kl_logstddevnormal_logstddevnormal)

  @mock.patch.dict(
      kl_module._DIVERGENCES,
      {(lsn.LogStddevNormal,
        lsn.LogStddevNormal): lambda *args, **kwargs: 42})
  def test_calling_custom_kl(self):
    # Check that the dispatch of tfp.kl_divergence actually goes to the
    # table we checked for above.
    dist_a = lsn.LogStddevNormal(jnp.array([0.0]), jnp.array([0.0]))
    dist_b = lsn.LogStddevNormal(jnp.array([0.0]), jnp.array([0.0]))
    self.assertEqual(tfp.distributions.kl_divergence(dist_a, dist_b), 42)
  # pylint:enable=protected-access

  def test_jitable(self):
    @jax.jit
    def jitted_function(event, dist):
      return dist.log_prob(event)

    dist = lsn.LogStddevNormal(np.array([0.0]), np.array([0.0]))
    event = dist.sample(seed=jax.random.PRNGKey(0))
    jitted_function(event, dist)

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    loc = jnp.array(np.random.randn(3, 4, 5))
    log_scale = jnp.array(np.random.randn(3, 4, 5))
    dist = lsn.LogStddevNormal(loc=loc, log_scale=log_scale)
    sliced_dist = dist[slice_]
    np.testing.assert_allclose(sliced_dist.mean(), loc[slice_], rtol=1e-3)
    np.testing.assert_allclose(
        sliced_dist.log_scale, log_scale[slice_], rtol=1e-3)
    self.assertIsInstance(sliced_dist, lsn.LogStddevNormal)

  def test_slice_different_parameterization(self):
    loc = jnp.array(np.random.randn(4))
    log_scale = jnp.array(np.random.randn(3, 4))
    dist = lsn.LogStddevNormal(loc=loc, log_scale=log_scale)
    np.testing.assert_allclose(
        dist[0].mean(), loc, rtol=1e-3)  # Not slicing loc.
    np.testing.assert_allclose(dist[0].log_scale, log_scale[0], rtol=1e-3)

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `masked_coupling.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import bijector as distrax_bijector
from distrax._src.bijectors import block
from distrax._src.bijectors import masked_coupling
import jax
import jax.numpy as jnp
import numpy as np


def _create_masked_coupling_bijector(event_shape, event_ndims=None):
  key = jax.random.PRNGKey(101)
  return masked_coupling.MaskedCoupling(
      mask=jax.random.choice(key, jnp.array([True, False]), event_shape),
      conditioner=lambda x: x**2,
      bijector=lambda _: lambda x: 2. * x + 3.,
      event_ndims=event_ndims)


class MaskedCouplingTest(parameterized.TestCase):

  def test_properties(self):
    bijector = _create_masked_coupling_bijector((4, 5), None)
    ones = jnp.ones((4, 5))
    np.testing.assert_allclose(bijector.conditioner(2 * ones), 4 * ones)
    assert callable(bijector.bijector(ones))
    self.assertEqual(bijector.mask.shape, (4, 5))

  @parameterized.named_parameters(
      ('jnp.float32', jnp.float32),
      ('jnp.int32', jnp.int32),
      ('jnp.float64', jnp.float64),
      ('jnp.int64', jnp.int64),
      ('jnp.complex64', jnp.complex64),
      ('jnp.complex128', jnp.complex128),
  )
  def test_raises_on_invalid_mask_dtype(self, dtype):
    with self.assertRaises(ValueError):
      masked_coupling.MaskedCoupling(
          mask=jnp.zeros((4,), dtype=dtype),
          conditioner=lambda x: x,
          bijector=lambda _: lambda x: x
      )

  @chex.all_variants
  @parameterized.parameters(
      {'event_ndims': None, 'batch_shape': (2, 3)},
      {'event_ndims': 0, 'batch_shape': (2, 3, 4, 5)},
      {'event_ndims': 1, 'batch_shape': (2, 3, 4)},
      {'event_ndims': 2, 'batch_shape': (2, 3)},
      {'event_ndims': 3, 'batch_shape': (2,)},
      {'event_ndims': 4, 'batch_shape': ()},
  )
  def test_shapes_are_correct(self, event_ndims, batch_shape):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bijector = _create_masked_coupling_bijector((4, 5), event_ndims)
    # Forward methods.
    y, logdet = self.variant(bijector.forward_and_log_det)(x)
    self.assertEqual(y.shape, (2, 3, 4, 5))
    self.assertEqual(logdet.shape, batch_shape)
    # Inverse methods.
    x, logdet = self.variant(bijector.inverse_and_log_det)(y)
    self.assertEqual(x.shape, (2, 3, 4, 5))
    self.assertEqual(logdet.shape, batch_shape)

  def test_non_default_inner_event_ndims(self):
    batch = 2
    event_shape = (7, 5, 3)
    inner_event_ndims = 1
    multipliers = jnp.array([4., 1., 0.5])

    class InnerBijector(distrax_bijector.Bijector):
      """A simple inner bijector."""

      def __init__(self):
        super().__init__(event_ndims_in=inner_event_ndims)

      def forward_and_log_det(self, x):
        return x * multipliers, jnp.full(x.shape[:-inner_event_ndims],
                                         jnp.sum(jnp.log(multipliers)))

      def inverse_and_log_det(self, y):
        return y / multipliers, jnp.full(x.shape[:-inner_event_ndims],
                                         -jnp.sum(jnp.log(multipliers)))

    bijector = masked_coupling.MaskedCoupling(
        mask=jnp.full(event_shape[:-inner_event_ndims], False),
        conditioner=lambda x: x,
        bijector=lambda _: InnerBijector(),
        inner_event_ndims=inner_event_ndims,
        event_ndims=len(event_shape))
    x = jnp.ones((batch,) + event_shape)
    # Test forward.
    y, ldj_y = bijector.forward_and_log_det(x)
    np.testing.assert_allclose(
        y, jnp.tile(multipliers[None, None, None, :], [batch, 7, 5, 1]))
    np.testing.assert_allclose(
        ldj_y,
        np.full([batch],
                np.prod(event_shape[:-inner_event_ndims]) *
                jnp.sum(jnp.log(multipliers))),
        rtol=1e-6)
    # Test inverse
    z, ldj_z = bijector.inverse_and_log_det(y)
    np.testing.assert_allclose(z, x)
    np.testing.assert_allclose(ldj_z, -ldj_y)

  @chex.all_variants
  def test_masking_works(self):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bijector = _create_masked_coupling_bijector((4, 5))
    mask = bijector.mask
    y = self.variant(bijector.forward)(x)
    np.testing.assert_array_equal(mask * y, mask * x)

  @chex.all_variants
  @parameterized.parameters(
      {'event_ndims': None},
      {'event_ndims': 0},
      {'event_ndims': 1},
      {'event_ndims': 2},
      {'event_ndims': 3},
      {'event_ndims': 4},
  )
  def test_inverse_methods_are_correct(self, event_ndims):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bijector = _create_masked_coupling_bijector((4, 5), event_ndims)
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(x)
    x_rec, logdet_inv = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(x_rec, x, atol=1e-6)
    np.testing.assert_allclose(logdet_fwd, -logdet_inv, atol=1e-6)

  @chex.all_variants
  @parameterized.parameters(
      {'event_ndims': None},
      {'event_ndims': 0},
      {'event_ndims': 1},
      {'event_ndims': 2},
      {'event_ndims': 3},
      {'event_ndims': 4},
  )
  def test_composite_methods_are_consistent(self, event_ndims):
    key = jax.random.PRNGKey(42)
    bijector = _create_masked_coupling_bijector((4, 5), event_ndims)
    # Forward methods.
    x = jax.random.normal(key, (2, 3, 4, 5))
    y1 = self.variant(bijector.forward)(x)
    logdet1 = self.variant(bijector.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bijector.forward_and_log_det)(x)
    np.testing.assert_allclose(y1, y2, atol=1e-8)
    np.testing.assert_allclose(logdet1, logdet2, atol=5e-6)
    # Inverse methods.
    y = jax.random.normal(key, (2, 3, 4, 5))
    x1 = self.variant(bijector.inverse)(y)
    logdet1 = self.variant(bijector.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bijector.inverse_and_log_det)(y)
    np.testing.assert_allclose(x1, x2, atol=1e-8)
    np.testing.assert_allclose(logdet1, logdet2, atol=5e-6)

  def test_raises_if_inner_bijector_is_not_scalar(self):
    key = jax.random.PRNGKey(101)
    event_shape = (2, 3)
    bijector = masked_coupling.MaskedCoupling(
        mask=jax.random.choice(key, jnp.array([True, False]), event_shape),
        conditioner=lambda x: x,
        bijector=lambda _: block.Block(lambda x: x, 1))
    with self.assertRaisesRegex(ValueError, r'match.*inner_event_ndims'):
      bijector.forward_and_log_det(jnp.zeros(event_shape))
    with self.assertRaisesRegex(ValueError, r'match.*inner_event_ndims'):
      bijector.inverse_and_log_det(jnp.zeros(event_shape))

  def test_raises_if_inner_bijector_has_wrong_inner_ndims(self):
    key = jax.random.PRNGKey(101)
    event_shape = (2, 3, 5)
    inner_event_ndims = 2
    bijector = masked_coupling.MaskedCoupling(
        mask=jax.random.choice(key, jnp.array([True, False]),
                               event_shape[:-inner_event_ndims]),
        event_ndims=len(event_shape),
        inner_event_ndims=inner_event_ndims,
        conditioner=lambda x: x,
        bijector=lambda _: block.Block(lambda x: x, 0))
    with self.assertRaisesRegex(ValueError, r'match.*inner_event_ndims'):
      bijector.forward_and_log_det(jnp.zeros(event_shape))
    with self.assertRaisesRegex(ValueError, r'match.*inner_event_ndims'):
      bijector.inverse_and_log_det(jnp.zeros(event_shape))

  def test_raises_on_invalid_input_shape(self):
    bij = _create_masked_coupling_bijector(event_shape=(2, 3))
    for fn in [bij.forward, bij.inverse,
               bij.forward_log_det_jacobian, bij.inverse_log_det_jacobian,
               bij.forward_and_log_det, bij.inverse_and_log_det]:
      with self.assertRaises(ValueError):
        fn(jnp.zeros((3,)))

  @chex.all_variants
  @parameterized.parameters(
      ((3, 4), (3, 4)),
      ((3, 4), (3, 1)),
      ((3, 4), (4,)),
      ((3, 4), ()),
      ((3, 1), (3, 4)),
      ((4,), (3, 4)),
      ((), (3, 4)),
  )
  def test_batched_mask(self, mask_batch_shape, input_batch_shape):
    def create_bijector(mask):
      return masked_coupling.MaskedCoupling(
          mask=mask,
          conditioner=lambda x: x**2,
          bijector=lambda _: lambda x: 2. * x + 3.,
          event_ndims=2)

    k1, k2 = jax.random.split(jax.random.PRNGKey(42))
    mask = jax.random.choice(
        k1, jnp.array([True, False]), mask_batch_shape + (5, 6))
    bijector = create_bijector(mask)

    x = jax.random.uniform(k2, input_batch_shape + (5, 6))
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(x)
    z, logdet_inv = self.variant(bijector.inverse_and_log_det)(x)

    output_batch_shape = jnp.broadcast_arrays(
        mask[..., 0, 0], x[..., 0, 0])[0].shape

    self.assertEqual(y.shape, output_batch_shape + (5, 6))
    self.assertEqual(z.shape, output_batch_shape + (5, 6))
    self.assertEqual(logdet_fwd.shape, output_batch_shape)
    self.assertEqual(logdet_inv.shape, output_batch_shape)

    mask = jnp.broadcast_to(
        mask, output_batch_shape + (5, 6)).reshape((-1, 5, 6))
    x = jnp.broadcast_to(x, output_batch_shape + (5, 6)).reshape((-1, 5, 6))
    y = y.reshape((-1, 5, 6))
    z = z.reshape((-1, 5, 6))
    logdet_fwd = logdet_fwd.flatten()
    logdet_inv = logdet_inv.flatten()

    for i in range(np.prod(output_batch_shape)):
      bijector = create_bijector(mask[i])
      this_y, this_logdet_fwd = self.variant(bijector.forward_and_log_det)(x[i])
      this_z, this_logdet_inv = self.variant(bijector.inverse_and_log_det)(x[i])
      np.testing.assert_allclose(this_y, y[i], atol=1e-7)
      np.testing.assert_allclose(this_z, z[i], atol=1e-7)
      np.testing.assert_allclose(this_logdet_fwd, logdet_fwd[i], atol=1e-5)
      np.testing.assert_allclose(this_logdet_inv, logdet_inv[i], atol=1e-5)

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = _create_masked_coupling_bijector((4, 5), event_ndims=None)
    x = np.zeros((2, 3, 4, 5))
    f(x, bijector)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `rational_quadratic_spline.py`.

Float64 is enabled in these tests. We keep them separate from other tests to
avoid interfering with types elsewhere.
"""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import rational_quadratic_spline
from jax import config as jax_config
import jax.numpy as jnp


def setUpModule():
  jax_config.update('jax_enable_x64', True)


class RationalQuadraticSplineFloat64Test(chex.TestCase):
  """Tests for rational quadratic spline that use float64."""

  def _assert_dtypes(self, bij, x, dtype):
    """Asserts dtypes."""
    # Sanity check to make sure float64 is enabled.
    x_64 = jnp.zeros([])
    self.assertEqual(jnp.float64, x_64.dtype)

    y, logd = self.variant(bij.forward_and_log_det)(x)
    self.assertEqual(dtype, y.dtype)
    self.assertEqual(dtype, logd.dtype)
    y, logd = self.variant(bij.inverse_and_log_det)(x)
    self.assertEqual(dtype, y.dtype)
    self.assertEqual(dtype, logd.dtype)

  @chex.all_variants
  @parameterized.product(
      dtypes=[(jnp.float32, jnp.float32, jnp.float32),
              (jnp.float32, jnp.float64, jnp.float64),
              (jnp.float64, jnp.float32, jnp.float64),
              (jnp.float64, jnp.float64, jnp.float64)],
      boundary_slopes=['unconstrained', 'lower_identity', 'upper_identity',
                       'identity', 'circular'])
  def test_dtypes(self, dtypes, boundary_slopes):
    x_dtype, params_dtype, result_dtype = dtypes
    x = jnp.zeros([3], x_dtype)
    self.assertEqual(x_dtype, x.dtype)
    spline = rational_quadratic_spline.RationalQuadraticSpline(
        jnp.zeros([25], params_dtype), 0., 1., boundary_slopes=boundary_slopes)
    self._assert_dtypes(spline, x, result_dtype)


if __name__ == '__main__':
  absltest.main()
